# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import json
import os
import shutil
from unittest import mock

import pandas as pd
import pytest
import torch

from ludwig.api import LudwigModel
from ludwig.callbacks import Callback
from ludwig.constants import BATCH_SIZE, ENCODER, TRAINER, TYPE
from ludwig.globals import MODEL_HYPERPARAMETERS_FILE_NAME
from ludwig.models.inference import InferenceModule
from ludwig.utils.data_utils import read_csv
from tests.integration_tests.utils import (
    category_feature,
    ENCODERS,
    generate_data,
    get_weights,
    image_feature,
    run_api_experiment,
    sequence_feature,
    text_feature,
)


def run_api_experiment_separated_datasets(input_features, output_features, data_csv):
    """Helper method to avoid code repetition in running an experiment.

    :param input_features: input schema
    :param output_features: output schema
    :param data_csv: path to data
    :return: None
    """
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }

    model = LudwigModel(config)

    # Training with dataframe
    data_df = read_csv(data_csv)
    train_df = data_df.sample(frac=0.8)
    test_df = data_df.drop(train_df.index).sample(frac=0.5)
    validation_df = data_df.drop(train_df.index).drop(test_df.index)

    basename, ext = os.path.splitext(data_csv)
    train_fname = basename + ".train" + ext
    val_fname = basename + ".validation" + ext
    test_fname = basename + ".test" + ext
    output_dirs = []

    try:
        train_df.to_csv(train_fname)
        validation_df.to_csv(val_fname)
        test_df.to_csv(test_fname)

        # Training with csv
        _, _, output_dir = model.train(
            training_set=train_fname,
            skip_save_processed_input=True,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
        )
        output_dirs.append(output_dir)

        _, _, output_dir = model.train(
            training_set=train_fname,
            validation_set=val_fname,
            skip_save_processed_input=True,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
        )
        output_dirs.append(output_dir)

        _, _, output_dir = model.train(
            training_set=train_fname,
            validation_set=val_fname,
            test_set=test_fname,
            skip_save_processed_input=True,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
        )
        output_dirs.append(output_dir)

        _, output_dir = model.predict(dataset=test_fname)
        output_dirs.append(output_dir)

    finally:
        # Remove results/intermediate data saved to disk
        os.remove(train_fname)
        os.remove(val_fname)
        os.remove(test_fname)
        for output_dir in output_dirs:
            shutil.rmtree(output_dir, ignore_errors=True)

    output_dirs = []
    try:
        _, _, output_dir = model.train(
            training_set=train_df,
            skip_save_processed_input=True,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
        )
        output_dirs.append(output_dir)

        _, _, output_dir = model.train(
            training_set=train_df,
            validation_set=validation_df,
            skip_save_processed_input=True,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
        )
        output_dirs.append(output_dir)

        _, _, output_dir = model.train(
            training_set=train_df,
            validation_set=validation_df,
            test_set=test_df,
            skip_save_processed_input=True,
            skip_save_progress=True,
            skip_save_unprocessed_output=True,
        )
        output_dirs.append(output_dir)

        _, output_dir = model.predict(dataset=data_df)
        output_dirs.append(output_dir)

    finally:
        for output_dir in output_dirs:
            shutil.rmtree(output_dir, ignore_errors=True)


def test_api_intent_classification(csv_filename):
    # Single sequence input, single category output
    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    for encoder in ENCODERS:
        input_features[0][ENCODER][TYPE] = encoder
        run_api_experiment(input_features, output_features, data_csv=rel_path)


def test_api_intent_classification_separated(csv_filename):
    # Single sequence input, single category output
    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    for encoder in ENCODERS:
        input_features[0][ENCODER][TYPE] = encoder
        run_api_experiment_separated_datasets(input_features, output_features, data_csv=rel_path)


def test_api_train_online(csv_filename):
    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]
    data_csv = generate_data(input_features, output_features, csv_filename)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
    }
    model = LudwigModel(config)

    for _ in range(2):
        model.train_online(dataset=data_csv)
    model.predict(dataset=data_csv)


def test_api_training_set(tmpdir):
    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]

    data_csv = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))
    val_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "validation.csv"))
    test_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "test.csv"))

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
    }
    model = LudwigModel(config)
    model.train(training_set=data_csv, validation_set=val_csv, test_set=test_csv)
    model.predict(dataset=test_csv)

    # Train again, this time the HDF5 cache will be used
    model.train(training_set=data_csv, validation_set=val_csv, test_set=test_csv)


def test_api_training_determinism(tmpdir):
    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]

    data_csv = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        "trainer": {BATCH_SIZE: 128},  # batch size must be fixed for determinism
    }

    # Train the model 3 times:
    #
    # 1. seed x
    # 2. seed y
    # 3. seed x
    #
    # Check that models (1) and (3) produce the same weights,
    # but (1) and (2) do not
    rand_x = 42
    rand_y = 24

    model_1 = LudwigModel(config)
    model_1.train(dataset=data_csv, output_directory=tmpdir, random_seed=rand_x)

    model_2 = LudwigModel(config)
    model_2.train(dataset=data_csv, output_directory=tmpdir, random_seed=rand_y)

    model_3 = LudwigModel(config)
    model_3.train(dataset=data_csv, output_directory=tmpdir, random_seed=rand_x)

    model_weights_1 = get_weights(model_1.model)
    model_weights_2 = get_weights(model_2.model)
    model_weights_3 = get_weights(model_3.model)

    divergence = False
    for weight_1, weight_2 in zip(model_weights_1, model_weights_2):
        if not torch.allclose(weight_1, weight_2):
            divergence = True
            break
    assert divergence, "model_1 and model_2 have identical weights with different seeds!"

    for weight_1, weight_3 in zip(model_weights_1, model_weights_3):
        assert torch.allclose(weight_1, weight_3)


def run_api_commands(
    input_features,
    output_features,
    data_csv,
    output_dir,
    skip_save_training_description=False,
    skip_save_training_statistics=False,
    skip_save_model=False,
    skip_save_progress=False,
    skip_save_log=False,
    skip_save_processed_input=False,
    skip_save_unprocessed_output=False,
    skip_save_predictions=False,
    skip_save_eval_stats=False,
    skip_collect_predictions=False,
    skip_collect_overall_stats=False,
):
    """Helper method to avoid code repetition in running an experiment.

    :param input_features: input schema
    :param output_features: output schema
    :param data_csv: path to data
    :return: None
    """
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }

    model = LudwigModel(config)

    # Training with csv
    model.train(
        dataset=data_csv,
        skip_save_training_description=skip_save_training_description,
        skip_save_training_statistics=skip_save_training_statistics,
        skip_save_model=skip_save_model,
        skip_save_progress=skip_save_progress,
        skip_save_log=skip_save_log,
        skip_save_processed_input=skip_save_processed_input,
        output_directory=output_dir,
    )
    model.predict(
        dataset=data_csv,
        skip_save_unprocessed_output=skip_save_unprocessed_output,
        skip_save_predictions=skip_save_predictions,
        output_directory=output_dir,
    )
    model.evaluate(
        dataset=data_csv,
        skip_save_unprocessed_output=skip_save_unprocessed_output,
        skip_save_predictions=skip_save_predictions,
        skip_save_eval_stats=skip_save_eval_stats,
        collect_predictions=not skip_collect_predictions,
        collect_overall_stats=not skip_collect_overall_stats,
        output_directory=output_dir,
    )
    model.experiment(
        dataset=data_csv,
        skip_save_training_description=skip_save_training_description,
        skip_save_training_statistics=skip_save_training_statistics,
        skip_save_model=skip_save_model,
        skip_save_progress=skip_save_progress,
        skip_save_log=skip_save_log,
        skip_save_processed_input=skip_save_processed_input,
        skip_save_unprocessed_output=skip_save_unprocessed_output,
        skip_save_predictions=skip_save_predictions,
        skip_save_eval_stats=skip_save_eval_stats,
        skip_collect_predictions=skip_collect_predictions,
        skip_collect_overall_stats=skip_collect_overall_stats,
        output_directory=output_dir,
    )


@pytest.mark.parametrize("skip_save_training_description", [False, True])
@pytest.mark.parametrize("skip_save_training_statistics", [False, True])
@pytest.mark.parametrize("skip_save_model", [False, True])
@pytest.mark.parametrize("skip_save_progress", [False, True])
@pytest.mark.parametrize("skip_save_log", [False, True])
@pytest.mark.parametrize("skip_save_processed_input", [False, True])
def test_api_skip_parameters_train(
    tmpdir,
    csv_filename,
    skip_save_training_description,
    skip_save_training_statistics,
    skip_save_model,
    skip_save_progress,
    skip_save_log,
    skip_save_processed_input,
):
    # Single sequence input, single category output
    input_features = [category_feature(encoder={"vocab_size": 5})]
    output_features = [category_feature(decoder={"vocab_size": 5})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, csv_filename))
    run_api_commands(
        input_features,
        output_features,
        data_csv=rel_path,
        output_dir=tmpdir,
        skip_save_training_description=skip_save_training_description,
        skip_save_training_statistics=skip_save_training_statistics,
        skip_save_model=skip_save_model,
        skip_save_progress=skip_save_progress,
        skip_save_log=skip_save_log,
        skip_save_processed_input=skip_save_processed_input,
    )


@pytest.mark.parametrize("skip_save_unprocessed_output", [False, True])
@pytest.mark.parametrize("skip_save_predictions", [False, True])
def test_api_skip_parameters_predict(
    tmpdir,
    csv_filename,
    skip_save_unprocessed_output,
    skip_save_predictions,
):
    # Single sequence input, single category output
    input_features = [category_feature(encoder={"vocab_size": 5})]
    output_features = [category_feature(decoder={"vocab_size": 5})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, csv_filename))
    run_api_commands(
        input_features,
        output_features,
        data_csv=rel_path,
        output_dir=tmpdir,
        skip_save_unprocessed_output=skip_save_unprocessed_output,
        skip_save_predictions=skip_save_predictions,
    )


@pytest.mark.parametrize("skip_save_unprocessed_output", [False, True])
@pytest.mark.parametrize("skip_save_predictions", [False, True])
@pytest.mark.parametrize("skip_save_eval_stats", [False, True])
@pytest.mark.parametrize("skip_collect_predictions", [False, True])
@pytest.mark.parametrize("skip_collect_overall_stats", [False, True])
def test_api_skip_parameters_evaluate(
    tmpdir,
    csv_filename,
    skip_save_unprocessed_output,
    skip_save_predictions,
    skip_save_eval_stats,
    skip_collect_predictions,
    skip_collect_overall_stats,
):
    # Single sequence input, single category output
    input_features = [category_feature(encoder={"vocab_size": 5})]
    output_features = [category_feature(decoder={"vocab_size": 5})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, csv_filename))
    run_api_commands(
        input_features,
        output_features,
        data_csv=rel_path,
        output_dir=tmpdir,
        skip_save_unprocessed_output=skip_save_unprocessed_output,
        skip_save_predictions=skip_save_predictions,
        skip_save_eval_stats=skip_save_eval_stats,
        skip_collect_predictions=skip_collect_predictions,
        skip_collect_overall_stats=skip_collect_overall_stats,
    )


@pytest.mark.parametrize("epochs", [1, 2])
@pytest.mark.parametrize("batch_size", [4, 8])
@pytest.mark.parametrize("num_examples", [16, 32])
@pytest.mark.parametrize("steps_per_checkpoint", [1, 2])
def test_api_callbacks(tmpdir, csv_filename, epochs, batch_size, num_examples, steps_per_checkpoint):
    mock_callback = mock.Mock(wraps=Callback())

    steps_per_epoch = num_examples / batch_size
    total_checkpoints = (steps_per_epoch / steps_per_checkpoint) * epochs
    total_batches = epochs * (num_examples / batch_size)

    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {
            "epochs": epochs,
            "batch_size": batch_size,
            "steps_per_checkpoint": steps_per_checkpoint,
            "early_stop": 0,  # Disable early stopping.
        },
    }
    model = LudwigModel(config, callbacks=[mock_callback])

    data_csv = generate_data(
        input_features, output_features, os.path.join(tmpdir, csv_filename), num_examples=num_examples
    )
    val_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "validation.csv"))
    test_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "test.csv"))

    model.train(training_set=data_csv, validation_set=val_csv, test_set=test_csv)

    assert mock_callback.on_epoch_start.call_count == epochs
    assert mock_callback.on_epoch_end.call_count == epochs

    assert mock_callback.should_early_stop.call_count == total_checkpoints

    assert mock_callback.on_validation_start.call_count == total_checkpoints
    assert mock_callback.on_validation_end.call_count == total_checkpoints

    assert mock_callback.on_test_start.call_count == total_checkpoints
    assert mock_callback.on_test_end.call_count == total_checkpoints

    assert mock_callback.on_batch_start.call_count == total_batches
    assert mock_callback.on_batch_end.call_count == total_batches

    assert mock_callback.on_eval_end.call_count == total_checkpoints
    assert mock_callback.on_eval_start.call_count == total_checkpoints


@pytest.mark.parametrize("epochs", [1, 2])
@pytest.mark.parametrize("batch_size", [4, 8])
@pytest.mark.parametrize("num_examples", [32, 64])
@pytest.mark.parametrize("checkpoints_per_epoch", [1, 2, 4])
def test_api_callbacks_checkpoints_per_epoch(
    tmpdir, csv_filename, epochs, batch_size, num_examples, checkpoints_per_epoch
):
    mock_callback = mock.Mock(wraps=Callback())

    total_checkpoints = epochs * checkpoints_per_epoch
    total_batches = epochs * (num_examples / batch_size)

    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {
            "epochs": epochs,
            "batch_size": batch_size,
            "checkpoints_per_epoch": checkpoints_per_epoch,
            "early_stop": 0,  # Disable early stopping.
        },
    }
    model = LudwigModel(config, callbacks=[mock_callback])

    data_csv = generate_data(
        input_features, output_features, os.path.join(tmpdir, csv_filename), num_examples=num_examples
    )
    val_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "validation.csv"))
    test_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "test.csv"))

    model.train(training_set=data_csv, validation_set=val_csv, test_set=test_csv)

    assert mock_callback.on_epoch_start.call_count == epochs
    assert mock_callback.on_epoch_end.call_count == epochs

    assert mock_callback.should_early_stop.call_count == total_checkpoints

    assert mock_callback.on_validation_start.call_count == total_checkpoints
    assert mock_callback.on_validation_end.call_count == total_checkpoints

    assert mock_callback.on_test_start.call_count == total_checkpoints
    assert mock_callback.on_test_end.call_count == total_checkpoints

    assert mock_callback.on_batch_start.call_count == total_batches
    assert mock_callback.on_batch_end.call_count == total_batches

    assert mock_callback.on_eval_end.call_count == total_checkpoints
    assert mock_callback.on_eval_start.call_count == total_checkpoints


def test_api_callbacks_default_train_steps(tmpdir, csv_filename):
    # Default for train_steps is -1: use epochs.
    train_steps = None
    epochs = 10
    batch_size = 8
    num_examples = 80
    mock_callback = mock.Mock(wraps=Callback())

    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": epochs, "train_steps": train_steps, "batch_size": batch_size},
    }
    model = LudwigModel(config, callbacks=[mock_callback])
    model.train(
        training_set=generate_data(
            input_features, output_features, os.path.join(tmpdir, csv_filename), num_examples=num_examples
        )
    )

    assert mock_callback.on_epoch_start.call_count == epochs


def test_api_callbacks_fixed_train_steps(tmpdir, csv_filename):
    # If train_steps is set manually, epochs is ignored.
    train_steps = 100
    epochs = 2
    batch_size = 8
    num_examples = 80
    mock_callback = mock.Mock(wraps=Callback())

    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": epochs, "train_steps": train_steps, "batch_size": batch_size},
    }
    model = LudwigModel(config, callbacks=[mock_callback])
    model.train(
        training_set=generate_data(
            input_features, output_features, os.path.join(tmpdir, csv_filename), num_examples=num_examples
        )
    )

    # There are 10 steps per epoch, so 100 train steps => 10 epochs.
    assert mock_callback.on_epoch_start.call_count == 10


def test_api_callbacks_fixed_train_steps_less_than_one_epoch(tmpdir, csv_filename):
    # If train_steps is set manually, epochs is ignored.
    train_steps = total_batches = 6
    steps_per_checkpoint = 2
    batch_size = 8
    num_examples = 80
    mock_callback = mock.Mock(wraps=Callback())

    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 5}, reduce_input="sum")]
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {
            "train_steps": train_steps,
            "steps_per_checkpoint": steps_per_checkpoint,
            "batch_size": batch_size,
        },
    }
    model = LudwigModel(config, callbacks=[mock_callback])
    model.train(
        training_set=generate_data(
            input_features, output_features, os.path.join(tmpdir, csv_filename), num_examples=num_examples
        )
    )

    assert mock_callback.on_epoch_start.call_count == 1
    assert mock_callback.on_epoch_end.call_count == 1
    # The total number of batches is the number of train_steps
    assert mock_callback.on_batch_end.call_count == total_batches
    # The total number of evals is the number of times checkpoints are made
    assert mock_callback.on_eval_end.call_count == train_steps // steps_per_checkpoint


def test_api_save_torchscript(tmpdir):
    """Tests successful saving and loading of model in TorchScript format."""
    input_features = [category_feature(encoder={"vocab_size": 5})]
    output_features = [
        category_feature(name="class", decoder={"vocab_size": 5}, reduce_input="sum", output_feature=True)
    ]

    data_csv = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))
    val_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "validation.csv"))
    test_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "test.csv"))

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
    }
    model = LudwigModel(config)
    model.train(training_set=data_csv, validation_set=val_csv, test_set=test_csv, output_directory=tmpdir)

    test_df = pd.read_csv(test_csv)
    output_df_expected, _ = model.predict(test_df, return_type=pd.DataFrame)

    save_path = os.path.join(tmpdir, "torchscript")
    os.makedirs(save_path, exist_ok=True)
    model.save_torchscript(save_path)
    inference_module = InferenceModule.from_directory(save_path)
    output_df, _ = inference_module.predict(test_df, return_type=pd.DataFrame)

    for col in output_df.columns:
        assert output_df[col].equals(output_df_expected[col])


def test_saved_weights_in_checkpoint(tmpdir):
    image_dest_folder = os.path.join(tmpdir, "generated_images")
    input_features = [
        text_feature(),
        image_feature(image_dest_folder),
    ]
    output_features = [category_feature(name="class", output_feature=True)]

    data_csv = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))
    val_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "validation.csv"))
    test_csv = shutil.copyfile(data_csv, os.path.join(tmpdir, "test.csv"))

    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {BATCH_SIZE: 128},
    }
    model = LudwigModel(config)
    _, _, output_dir = model.train(
        training_set=data_csv, validation_set=val_csv, test_set=test_csv, output_directory=tmpdir
    )

    config_save_path = os.path.join(output_dir, "model", MODEL_HYPERPARAMETERS_FILE_NAME)
    with open(config_save_path) as f:
        saved_config = json.load(f)
    saved_input_features = saved_config["input_features"]
    for saved_input_feature in saved_input_features:
        assert "encoder" in saved_input_feature
        input_feature_encoder = saved_input_feature["encoder"]
        assert "saved_weights_in_checkpoint" in input_feature_encoder
        assert input_feature_encoder["saved_weights_in_checkpoint"]



import pytest
import torch

from ludwig.modules import metric_modules


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(6).float()])
def test_rmse_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.RMSEMetric()
    metric.update(preds, target)
    assert output == metric.compute()


@pytest.mark.parametrize("preds", [torch.tensor([0.2, 0.3, 0.8, 0.1])])
@pytest.mark.parametrize("target", [torch.tensor([0, 0, 1, 1])])
@pytest.mark.parametrize("output", [torch.tensor(0.5)])
def test_roc_auc_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.BinaryAUROCMetric(task="binary")
    metric.update(preds, target)
    assert output == metric.compute()


@pytest.mark.parametrize("preds", [torch.tensor([0.2, 0.3, 0.8, 0.1, 0.8])])
@pytest.mark.parametrize("target", [torch.tensor([0, 0, 1, 1, 0])])
@pytest.mark.parametrize("output", [torch.tensor(0.6667).float()])
def test_specificity_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.SpecificityMetric()
    metric.update(preds, target)
    assert torch.isclose(output, metric.compute(), rtol=0.0001)


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(0.7527).float()])
def test_rmspe_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.RMSPEMetric()
    metric.update(preds, target)
    assert torch.isclose(output, metric.compute(), rtol=0.0001)


@pytest.mark.parametrize(
    "preds,target,num_outputs,output",
    [
        (torch.arange(3), torch.arange(3, 6), 1, torch.tensor(-12.5)),
        (torch.arange(6).reshape(3, 2), torch.arange(6, 12).reshape(3, 2), 2, torch.tensor(-12.5)),
    ],
)
def test_r2_score(preds: torch.Tensor, target: torch.Tensor, num_outputs: int, output: torch.Tensor):
    metric = metric_modules.R2Score(num_outputs=num_outputs)
    metric.update(preds, target)
    assert metric.compute() == output


def test_r2_score_single_sample():
    metric = metric_modules.R2Score(num_outputs=1)
    metric.update(preds=torch.tensor([0.8]), target=torch.arange(1))
    assert torch.isnan(metric.compute())


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(-21.4655).float()])
def test_bwcewl_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.BWCEWLMetric()
    metric.update(preds, target)
    assert torch.isclose(output, metric.compute(), rtol=0.0001)


@pytest.mark.parametrize("preds", [torch.tensor([[0.5, 0.5], [0.2, 0.8], [0.6, 0.4]])])
@pytest.mark.parametrize("target", [torch.tensor([1, 1, 0])])
@pytest.mark.parametrize("output", [torch.tensor(0.5763)])
def test_softmax_cross_entropy_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.SoftmaxCrossEntropyMetric()
    metric.update(preds, target)
    assert torch.isclose(output, metric.compute(), rtol=0.0001)


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(-21.4655).float()])
def test_sigmoid_cross_entropy_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.SigmoidCrossEntropyMetric()
    metric.update(preds, target)
    assert torch.isclose(output, metric.compute(), rtol=0.0001)


@pytest.mark.parametrize(
    "preds,target,output",
    [
        (
            torch.tensor([[0, 1], [3, 2], [4, 5]]),
            torch.tensor([[0, 1], [1, 2], [4, 5]]),
            torch.tensor(0.8),
        ),
        (
            torch.tensor([[0, 1, 2], [1, 3, 4], [3, 4, 5]]),
            torch.tensor([[0, 1, 2], [1, 1, 4], [3, 4, 5]]),
            torch.tensor(0.8750),
        ),
        (
            torch.tensor([[1, 5, 1, 5, 1, 5, 12, 12, 12], [10, 1, 5, 1, 5, 12, 12, 12, 12]]),
            torch.tensor([[1, 9, 5, 7, 5, 9, 13, 6, 0], [1, 9, 7, 13, 4, 7, 7, 7, 0]]),
            torch.tensor(0.05555555),
        ),
    ],
)
def test_token_accuracy_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.TokenAccuracyMetric()
    metric.update(preds, target)
    assert torch.allclose(metric.compute(), output)


def test_sequence_accuracy_metric():
    target = torch.tensor(
        [
            [1, 6, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 6, 5, 4, 0],
            [1, 6, 5, 4, 0],
            [1, 6, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 6, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 6, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
        ]
    )
    preds = torch.tensor(
        [
            [1, 6, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 6, 5, 4, 0],
            [1, 6, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 6, 5, 4, 0],
            [1, 6, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 6, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 6, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
            [1, 4, 5, 4, 0],
        ]
    )
    metric = metric_modules.SequenceAccuracyMetric()
    metric.update(preds, target)
    assert torch.isclose(metric.compute(), torch.tensor(0.8438), rtol=0.0001)


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2)])
@pytest.mark.parametrize("target", [torch.tensor([[0, 1], [2, 1], [4, 5]]).float()])
@pytest.mark.parametrize("output", [torch.tensor(0.7500).float()])
def test_category_accuracy(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.CategoryAccuracy(num_classes=6)
    metric.update(preds, target)
    assert torch.isclose(output, metric.compute(), rtol=0.0001)


@pytest.mark.parametrize(
    "preds,target,output,k",
    [
        (
            torch.tensor([[0.1, 0.9, 0], [0.3, 0.1, 0.6], [0.2, 0.5, 0.3]]),
            torch.tensor([0, 1, 2]),
            torch.tensor(0.6667).float(),
            2,
        )
    ],
)
def test_hits_at_k_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor, k: int):
    metric = metric_modules.HitsAtKMetric(num_classes=3, top_k=k)
    metric.update(preds, target)
    assert torch.isclose(output, metric.compute(), rtol=0.0001)


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(6).float()])
def test_mae_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.MAEMetric()
    metric.update(preds, target)
    assert output == metric.compute()


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(36).float()])
def test_mse_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.MSEMetric()
    metric.update(preds, target)
    assert output == metric.compute()


@pytest.mark.parametrize("preds", [torch.tensor([[0, 1], [1, 1]])])
@pytest.mark.parametrize("target", [torch.tensor([[1, 0], [1, 1]])])
@pytest.mark.parametrize("output", [torch.tensor(0.5)])
def test_jaccard_metric(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    metric = metric_modules.JaccardMetric()
    metric.update(preds, target)
    assert output == metric.compute()



import numpy as np
import pandas as pd
import pytest

from ludwig.backend.base import LocalBackend
from ludwig.constants import BALANCE_PERCENTAGE_TOLERANCE, NAME
from ludwig.data.preprocessing import balance_data


@pytest.mark.parametrize(
    "method, balance",
    [
        ("oversample_minority", 0.25),
        ("oversample_minority", 0.5),
        ("oversample_minority", 0.75),
        ("undersample_majority", 0.25),
        ("undersample_majority", 0.5),
        ("undersample_majority", 0.75),
        ("undersample_majority", 0.9),
    ],
)
def test_balance(method, balance):
    config = {
        "input_features": [
            {"name": "Index", "proc_column": "Index", "type": "number"},
            {"name": "random_1", "proc_column": "random_1", "type": "number"},
            {"name": "random_2", "proc_column": "random_2", "type": "number"},
        ],
        "output_features": [{"name": "Label", "proc_column": "Label", "type": "binary"}],
        "preprocessing": {"oversample_minority": None, "undersample_majority": None},
    }
    input_df = pd.DataFrame(
        {
            "Index": np.arange(0, 200, 1),
            "random_1": np.random.randint(0, 50, 200),
            "random_2": np.random.choice(["Type A", "Type B", "Type C", "Type D"], 200),
            "Label": np.concatenate((np.zeros(180), np.ones(20))),
            "split": np.zeros(200),
        }
    )

    config["preprocessing"][method] = balance
    backend = LocalBackend()

    test_df = balance_data(input_df, config["output_features"], config["preprocessing"], backend, 42)
    target = config["output_features"][0][NAME]
    majority_class = test_df[target].value_counts()[test_df[target].value_counts().idxmax()]
    minority_class = test_df[target].value_counts()[test_df[target].value_counts().idxmin()]
    new_class_balance = round(minority_class / majority_class, 2)

    assert abs(balance - new_class_balance) < BALANCE_PERCENTAGE_TOLERANCE



import logging
from typing import Dict, Tuple

from ludwig.constants import (
    BATCH_SIZE,
    CATEGORY,
    COMBINER,
    DECODER,
    DEFAULTS,
    ENCODER,
    EPOCHS,
    FILL_WITH_CONST,
    INPUT_FEATURES,
    LOSS,
    OUTPUT_FEATURES,
    PREPROCESSING,
    TEXT,
    TRAINER,
    TYPE,
)
from ludwig.schema.model_config import ModelConfig
from tests.integration_tests.utils import category_feature, generate_data, run_experiment, text_feature

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.getLogger("ludwig").setLevel(logging.INFO)


def _prepare_data(csv_filename: str) -> Tuple[Dict, str]:
    input_features = [
        text_feature(name="title", reduce_output="sum"),
        text_feature(name="summary"),
        category_feature(vocab_size=3),
        category_feature(vocab_size=3),
    ]

    output_features = [text_feature(name="article", embedding_size=3, output_feature=True)]

    dataset = generate_data(input_features, output_features, csv_filename)

    config = {
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        COMBINER: {TYPE: "concat", "num_fc_layers": 2},
        TRAINER: {EPOCHS: 1, "learning_rate": 0.001, BATCH_SIZE: 128},
        DEFAULTS: {
            CATEGORY: {
                PREPROCESSING: {"missing_value_strategy": FILL_WITH_CONST, "fill_value": "<CUSTOM_TOK>"},
                ENCODER: {TYPE: "sparse"},
                DECODER: {"norm_params": None, "dropout": 0.1, "use_bias": True},
            },
            TEXT: {
                PREPROCESSING: {"most_common": 10, "padding_symbol": "<PADDING>"},
                ENCODER: {TYPE: "rnn"},
                DECODER: {TYPE: "generator", "num_fc_layers": 2, "dropout": 0.1},
                LOSS: {"confidence_penalty": 0.1},
            },
        },
    }

    return config, dataset


def test_run_experiment_with_global_default_parameters(csv_filename):
    config, dataset = _prepare_data(csv_filename)

    run_experiment(config=config, dataset=dataset)


def test_global_defaults_with_encoder_dependencies():
    input_features = [text_feature(name="title", reduce_output="sum")]
    output_features = [category_feature(name="article", embedding_size=3, output_feature=True)]
    del input_features[0][ENCODER]

    config = {
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        DEFAULTS: {
            TEXT: {
                ENCODER: {TYPE: "bert"},
            }
        },
    }

    # Config should populate with the additional required fields for bert
    updated_config = ModelConfig.from_dict(config).to_dict()

    assert updated_config[INPUT_FEATURES][0][ENCODER][TYPE] == "bert"
    assert updated_config[INPUT_FEATURES][0][ENCODER]["pretrained_model_name_or_path"] == "bert-base-uncased"



from typing import Any, Dict, Optional

import pytest

from ludwig.constants import DEFAULTS, ENCODER, TEXT, TRAINABLE, TRAINER, TYPE
from ludwig.schema.model_config import ModelConfig
from ludwig.utils.heuristics import get_auto_learning_rate


@pytest.mark.parametrize(
    "text_encoder,expected_lr",
    [
        (None, 0.001),
        ({}, 0.00001),
        ({"type": "parallel_cnn"}, 0.0001),
        ({"type": "bert"}, 0.00002),
        ({"type": "bert", "trainable": True}, 0.00001),
        ({"type": "bert", "trainable": True, "use_pretrained": False}, 0.0001),
    ],
    ids=["no_text", "default_electra", "parallel_cnn", "bert_fixed", "bert_trainable", "bert_untrained"],
)
def test_get_auto_learning_rate(text_encoder: Optional[Dict[str, Any]], expected_lr: float):
    input_features = [{"name": "bin1", "type": "binary"}]
    if text_encoder is not None:
        input_features.append({"name": "text1", "type": "text", "encoder": text_encoder})

    config = {
        "input_features": input_features,
        "output_features": [{"name": "bin2", "type": "binary"}],
        TRAINER: {
            "train_steps": 1,
            "learning_rate": "auto",
        },
        DEFAULTS: {
            TEXT: {
                ENCODER: {
                    # Note that encoder defaults are all or nothing: if the encoder type is overridden, trainable
                    # here is ignored
                    TYPE: "electra",
                    TRAINABLE: True,
                }
            }
        },
    }

    config = ModelConfig.from_dict(config)
    lr = get_auto_learning_rate(config)
    assert lr == expected_lr



from copy import deepcopy
from typing import Dict

import pytest
import torch

from ludwig.constants import ENCODER, TYPE
from ludwig.features.category_feature import CategoryInputFeature
from ludwig.schema.features.category_feature import ECDCategoryInputFeatureConfig
from ludwig.schema.utils import load_config_with_kwargs
from ludwig.utils.misc_utils import merge_dict
from ludwig.utils.torch_utils import get_torch_device

BATCH_SIZE = 2
DEVICE = get_torch_device()


@pytest.fixture(scope="module")
def category_config():
    return {
        "name": "category_column_name",
        "type": "category",
        "tied": None,
        "encoder": {
            "embedding_size": 256,
            "embeddings_on_cpu": False,
            "pretrained_embeddings": None,
            "embeddings_trainable": True,
            "dropout": 0.0,
            "vocab": ["a", "b", "c"],
            "embedding_initializer": None,
        },
    }


@pytest.mark.parametrize("encoder", ["dense", "sparse"])
def test_category_input_feature(
    category_config: Dict,
    encoder: str,
) -> None:
    # setup image input feature definition
    category_def = deepcopy(category_config)
    category_def[ENCODER][TYPE] = encoder

    # pickup any other missing parameters
    defaults = ECDCategoryInputFeatureConfig().to_dict()
    category_def = merge_dict(defaults, category_def)

    # ensure no exceptions raised during build
    category_config, _ = load_config_with_kwargs(ECDCategoryInputFeatureConfig, category_def)
    input_feature_obj = CategoryInputFeature(category_config).to(DEVICE)

    # check one forward pass through input feature
    input_tensor = torch.randint(0, 3, size=(BATCH_SIZE,), dtype=torch.int32).to(DEVICE)

    encoder_output = input_feature_obj(input_tensor)
    assert encoder_output["encoder_output"].shape == (BATCH_SIZE, *input_feature_obj.output_shape)



from os import PathLike
from typing import Optional, Type, Union

from transformers import AutoTokenizer
from transformers.modeling_utils import PreTrainedModel
from transformers.tokenization_utils import PreTrainedTokenizer

from ludwig.utils.error_handling_utils import default_retry


@default_retry()
def load_pretrained_hf_model(
    modelClass: Type,
    pretrained_model_name_or_path: Optional[Union[str, PathLike]],
    **pretrained_kwargs,
) -> PreTrainedModel:
    """Download a HuggingFace model.

    Downloads a model from the HuggingFace zoo with retry on failure.
    Args:
        modelClass: Class of the model to download.
        pretrained_model_name_or_path: Name of the model to download.
        pretrained_kwargs: Additional arguments to pass to the model constructor.
    Returns:
        The pretrained model object.
    """
    return modelClass.from_pretrained(pretrained_model_name_or_path, **pretrained_kwargs)


@default_retry()
def load_pretrained_hf_tokenizer(
    pretrained_model_name_or_path: Optional[Union[str, PathLike]], **pretrained_kwargs
) -> PreTrainedTokenizer:
    """Download a HuggingFace tokenizer.

    Args:
        pretrained_model_name_or_path: Name of the tokenizer to download.
        pretrained_kwargs: Additional arguments to pass to the tokenizer constructor.
    Returns:
        The pretrained tokenizer object.
    """
    return AutoTokenizer.from_pretrained(pretrained_model_name_or_path, **pretrained_kwargs)



import os
from random import choice
from string import ascii_lowercase, ascii_uppercase, digits

import pandas as pd
import pytest
import torch

from ludwig.backend import LOCAL_BACKEND
from ludwig.constants import BFILL, PROC_COLUMN
from ludwig.features.audio_feature import AudioFeatureMixin, AudioInputFeature
from ludwig.schema.features.audio_feature import AudioInputFeatureConfig
from ludwig.schema.utils import load_config_with_kwargs
from ludwig.utils.torch_utils import get_torch_device
from tests.integration_tests.utils import audio_feature, category_feature, generate_data

BATCH_SIZE = 2
SEQ_SIZE = 20
AUDIO_W_SIZE = 16

CHARS = ascii_uppercase + ascii_lowercase + digits
VOCAB = ["".join(choice(CHARS) for _ in range(2)) for _ in range(256)]
DEVICE = get_torch_device()


@pytest.mark.parametrize("encoder", ["rnn", "stacked_cnn", "parallel_cnn", "stacked_parallel_cnn", "rnn", "cnnrnn"])
def test_audio_input_feature(encoder: str) -> None:
    audio_config = {
        "name": "audio_feature",
        "type": "audio",
        "preprocessing": {
            "type": "fbank",
            "window_length_in_s": 0.04,
            "window_shift_in_s": 0.02,
            "num_filter_bands": 80,
            "audio_file_length_limit_in_s": 3.0,
        },
        "encoder": {
            "type": encoder,
            "should_embed": False,
            "vocab": VOCAB,
            "max_sequence_length": SEQ_SIZE,
            "embedding_size": AUDIO_W_SIZE,
        },
    }

    audio_config, _ = load_config_with_kwargs(AudioInputFeatureConfig, audio_config)
    audio_input_feature = AudioInputFeature(audio_config)

    audio_tensor = torch.randn([BATCH_SIZE, SEQ_SIZE, AUDIO_W_SIZE], dtype=torch.float32).to(DEVICE)
    encoder_output = audio_input_feature(audio_tensor)
    assert encoder_output["encoder_output"].shape[1:] == audio_input_feature.output_shape


@pytest.mark.parametrize("feature_type", ["raw", "stft", "stft_phase", "group_delay", "fbank"])
def test_add_feature_data(feature_type, tmpdir):
    preprocessing_params = {
        "audio_file_length_limit_in_s": 3.0,
        "missing_value_strategy": BFILL,
        "in_memory": True,
        "padding_value": 0,
        "norm": "per_file",
        "type": feature_type,
        "window_length_in_s": 0.04,
        "window_shift_in_s": 0.02,
        "num_fft_points": None,
        "window_type": "hamming",
        "num_filter_bands": 80,
    }
    audio_dest_folder = os.path.join(tmpdir, "generated_audio")
    audio_feature_config = audio_feature(audio_dest_folder, preprocessing=preprocessing_params)
    data_df_path = generate_data(
        [audio_feature_config],
        [category_feature(vocab_size=5, reduce_input="sum")],
        os.path.join(tmpdir, "data.csv"),
        num_examples=10,
    )
    data_df = pd.read_csv(data_df_path)
    metadata = {
        audio_feature_config["name"]: AudioFeatureMixin.get_feature_meta(
            data_df[audio_feature_config["name"]], preprocessing_params, LOCAL_BACKEND, True
        )
    }

    proc_df = {}
    AudioFeatureMixin.add_feature_data(
        feature_config=audio_feature_config,
        input_df=data_df,
        proc_df=proc_df,
        metadata=metadata,
        preprocessing_parameters=preprocessing_params,
        backend=LOCAL_BACKEND,
        skip_save_processed_input=False,
    )

    assert len(proc_df[audio_feature_config[PROC_COLUMN]]) == 10



import logging
import math
from typing import Any, Dict

from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR, ReduceLROnPlateau

from ludwig.constants import MINIMIZE, TRAINING, VALIDATION
from ludwig.modules.metric_registry import get_metric_objective
from ludwig.schema.lr_scheduler import LRSchedulerConfig
from ludwig.utils.metric_utils import TrainerMetric
from ludwig.utils.trainer_utils import ProgressTracker


class ReduceLROnPLateauCappedDecreases(ReduceLROnPlateau):
    def __init__(self, optimizer: Optimizer, mode: str, reduce_limit: int, factor: float, patience: int):
        super().__init__(optimizer, mode=mode, factor=factor, patience=patience)
        self.reduce_limit = reduce_limit
        self._num_reduce_lr = 0

    def step(self, metrics):
        if self._num_reduce_lr >= self.reduce_limit:
            # Already reduced the LR as many times as we will allow
            return

        return super().step(metrics)

    @property
    def num_reduce_lr(self) -> int:
        return self._num_reduce_lr

    def _reduce_lr(self, epoch):
        self._num_reduce_lr += 1
        self.apply_lr(epoch)

    def apply_lr(self, epoch=None):
        if self._num_reduce_lr == 0:
            return

        for i, param_group in enumerate(self.optimizer.param_groups):
            old_lr = float(param_group["lr"])
            new_lr = max(old_lr * math.pow(self.factor, self._num_reduce_lr), self.min_lrs[i])
            if old_lr - new_lr > self.eps:
                param_group["lr"] = new_lr
                if self.verbose:
                    epoch_str = ("%.2f" if isinstance(epoch, float) else "%.5d") % epoch
                    print("Epoch {}: reducing learning rate" " of group {} to {:.4e}.".format(epoch_str, i, new_lr))


class LRScheduler:
    def __init__(
        self,
        config: LRSchedulerConfig,
        optimizer: Optimizer,
        steps_per_checkpoint: int = 1000,
        total_steps: int = 10000,
    ):
        self.config = config
        self.optimizer = optimizer

        # Scheduler updated each training step
        self.step_info = StepInfo(steps_per_checkpoint, total_steps, self.config)
        self._train_scheduler = get_schedule_with_warmup(self.config, self.optimizer, self.step_info)

        # Scheduler updated each eval step
        self._eval_scheduler = None
        if self.config.reduce_on_plateau > 0:
            mode = "min" if get_metric_objective(self.config.reduce_eval_metric) == MINIMIZE else "max"
            self._eval_scheduler = ReduceLROnPLateauCappedDecreases(
                optimizer=self.optimizer,
                mode=mode,
                reduce_limit=self.config.reduce_on_plateau,
                factor=self.config.reduce_on_plateau_rate,
                patience=self.config.reduce_on_plateau_patience,
            )

        self.reset(steps_per_checkpoint, total_steps)

    def reset(self, steps_per_checkpoint: int, total_steps: int):
        # Retain state but update number of steps for training
        self.step_info.reset(steps_per_checkpoint, total_steps)

    def step(self):
        self._train_scheduler.step()

        if self._eval_scheduler is not None:
            # We apply this scheduler every eval step, not train step, so we don't want to call step() here.
            # However, we need to re-apply the LR reduction to the LR from the train scheduler, as the first scheduler
            # resets the LR back to the base LR.
            self._eval_scheduler.apply_lr()

    def eval_step(self, progress_tracker: ProgressTracker, validation_field: str):
        if self._eval_scheduler is None:
            # No reduce on plateau
            return

        if self.config.reduce_eval_split == TRAINING:
            split_metrics = progress_tracker.train_metrics
        elif self.config.reduce_eval_split == VALIDATION:
            split_metrics = progress_tracker.validation_metrics
        else:  # if self.config.reduce_eval_split == TEST:
            split_metrics = progress_tracker.test_metrics

        validation_metric = self.config.reduce_eval_metric
        last_metric: TrainerMetric = split_metrics[validation_field][validation_metric][-1]
        last_metric_value = last_metric[-1]

        prev_num_reductions = self._eval_scheduler.num_reduce_lr
        self._eval_scheduler.step(last_metric_value)

        num_reductions = self._eval_scheduler.num_reduce_lr
        if num_reductions > prev_num_reductions:
            # LR reduction -> update progress tracker
            progress_tracker.last_learning_rate_reduction_steps = progress_tracker.steps
            progress_tracker.last_learning_rate_reduction = 0
            progress_tracker.num_reductions_learning_rate += 1
        else:
            progress_tracker.last_learning_rate_reduction = (
                progress_tracker.steps - progress_tracker.last_learning_rate_reduction_steps
            )

    def state_dict(self) -> Dict[str, Any]:
        return {
            "train_scheduler_state": self._train_scheduler.state_dict(),
            "eval_scheduler_state": self._eval_scheduler.state_dict() if self._eval_scheduler is not None else {},
        }

    def load_state_dict(self, d: Dict[str, Any]):
        self._train_scheduler.load_state_dict(d["train_scheduler_state"])
        if self._eval_scheduler is not None:
            self._eval_scheduler.load_state_dict(d["eval_scheduler_state"])


class StepInfo:
    """Stores the steps_per_checkpoint and total_steps used during the current training run.

    This class is needed by LambdaLR to allow us to update the steps on training init without resetting the entire
    LRScheduler from scratch (which would result in resetting the optimizer learning rate).
    """

    def __init__(self, steps_per_checkpoint: int, total_steps: int, config: LRSchedulerConfig):
        self.config = config
        self.reset(steps_per_checkpoint, total_steps)

    def reset(self, steps_per_checkpoint: int, total_steps: int):
        self.steps_per_checkpoint = steps_per_checkpoint
        self.num_training_steps = total_steps

        if self.config.warmup_fraction > 0 and self.config.warmup_evaluations > 0:
            logging.info(
                "Both `learning_rate_scheduler.warmup_fraction` and `learning_rate_scheduler.warmup_evaluations` "
                "provided. The larger of the two (as a function of the total training steps) will be used."
            )

        num_warmup_steps = 0
        if self.config.warmup_fraction > 0:
            num_warmup_steps = max(self.config.warmup_fraction * self.num_training_steps, num_warmup_steps)
        if self.config.warmup_evaluations > 0:
            num_warmup_steps = max(self.config.warmup_evaluations * self.steps_per_checkpoint, num_warmup_steps)
        self.num_warmup_steps = num_warmup_steps


def get_schedule_with_warmup(
    config: LRSchedulerConfig,
    optimizer: Optimizer,
    step_info: StepInfo,
) -> LambdaLR:
    """Creates a learning rate scheduler that updates each training step."""
    decay_fn = decay_registry[config.decay]

    def lr_lambda(current_step: int):
        if current_step < step_info.num_warmup_steps:
            return float(current_step) / float(max(1, step_info.num_warmup_steps))
        return decay_fn(current_step, step_info.num_training_steps, step_info.num_warmup_steps, config)

    return LambdaLR(optimizer, lr_lambda, last_epoch=-1)


def no_decay(current_step: int, num_training_steps: int, num_warmup_steps: int, config: LRSchedulerConfig):
    return 1.0


def linear_decay(current_step: int, num_training_steps: int, num_warmup_steps: int, config: LRSchedulerConfig):
    return max(0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps)))


def exponential_decay(current_step: int, num_training_steps: int, num_warmup_steps: int, config: LRSchedulerConfig):
    decay_rate = float(config.decay_rate)
    decay_steps = float(config.decay_steps)
    step = float(current_step)
    exponent = 1 + step / decay_steps
    if config.staircase:
        exponent = math.ceil(exponent)
    return math.pow(decay_rate, exponent)


decay_registry = {
    None: no_decay,
    "linear": linear_decay,
    "exponential": exponential_decay,
}



import numpy as np
import pytest

from ludwig.utils.numerical_test_utils import assert_all_finite


@pytest.fixture
def finite_valued_dict():
    return {
        "scalar": 1,
        "metrics": {"val": 0.2, "series": [0.1, 0.2, 0.3], "ndarray": np.ones((8, 4, 2))},
    }


def test_assert_all_finite(finite_valued_dict):
    assert_all_finite(finite_valued_dict)


def test_fail_with_nan(finite_valued_dict):
    finite_valued_dict["scalar"] = float("nan")
    with pytest.raises(Exception):
        assert_all_finite(finite_valued_dict)


def test_fail_with_inf(finite_valued_dict):
    finite_valued_dict["scalar"] = float("inf")
    with pytest.raises(Exception):
        assert_all_finite(finite_valued_dict)


def test_fail_with_nan_in_list(finite_valued_dict):
    finite_valued_dict["scalar"] = float("nan")
    with pytest.raises(Exception):
        assert_all_finite(finite_valued_dict)


def test_fail_with_nan_in_ndarray(finite_valued_dict):
    finite_valued_dict["metrics"]["ndarray"][0, 0, 1] = np.nan
    with pytest.raises(Exception):
        assert_all_finite(finite_valued_dict)



from typing import Type

from ludwig.distributed.base import DistributedStrategy, LocalStrategy


def load_ddp():
    from ludwig.distributed.ddp import DDPStrategy

    return DDPStrategy


def load_fsdp():
    from ludwig.distributed.fsdp import FSDPStrategy

    return FSDPStrategy


def load_horovod():
    from ludwig.distributed.horovod import HorovodStrategy

    return HorovodStrategy


def load_local():
    return LocalStrategy


STRATEGIES = {"ddp": load_ddp, "fsdp": load_fsdp, "horovod": load_horovod, "local": load_local}


def get_current_dist_strategy(allow_local=True) -> Type[DistributedStrategy]:
    for strategy_loader in STRATEGIES.values():
        try:
            strategy_cls = strategy_loader()
        except ImportError:
            continue

        if strategy_cls.is_available():
            return strategy_cls

    if allow_local:
        return LocalStrategy

    raise RuntimeError("Expected current distributed strategy, but none is available")


def get_dist_strategy(name: str) -> Type[DistributedStrategy]:
    return STRATEGIES[name]()


def get_default_strategy_name() -> str:
    try:
        # Use horovod by default for now if it's available
        load_horovod()
        return "horovod"
    except ImportError:
        pass

    # Fallback to DDP if not
    return "ddp"



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, TYPE_CHECKING

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import (
    AUDIO,
    BAG,
    BINARY,
    CATEGORY,
    DATE,
    H3,
    IMAGE,
    NUMBER,
    SEQUENCE,
    SET,
    TEXT,
    TIMESERIES,
    VECTOR,
)
from ludwig.features.audio_feature import AudioFeatureMixin, AudioInputFeature
from ludwig.features.bag_feature import BagFeatureMixin, BagInputFeature
from ludwig.features.binary_feature import BinaryFeatureMixin, BinaryInputFeature, BinaryOutputFeature
from ludwig.features.category_feature import CategoryFeatureMixin, CategoryInputFeature, CategoryOutputFeature
from ludwig.features.date_feature import DateFeatureMixin, DateInputFeature
from ludwig.features.h3_feature import H3FeatureMixin, H3InputFeature
from ludwig.features.image_feature import ImageFeatureMixin, ImageInputFeature
from ludwig.features.number_feature import NumberFeatureMixin, NumberInputFeature, NumberOutputFeature
from ludwig.features.sequence_feature import SequenceFeatureMixin, SequenceInputFeature, SequenceOutputFeature
from ludwig.features.set_feature import SetFeatureMixin, SetInputFeature, SetOutputFeature
from ludwig.features.text_feature import TextFeatureMixin, TextInputFeature, TextOutputFeature
from ludwig.features.timeseries_feature import TimeseriesFeatureMixin, TimeseriesInputFeature
from ludwig.features.vector_feature import VectorFeatureMixin, VectorInputFeature, VectorOutputFeature
from ludwig.utils.misc_utils import get_from_registry

if TYPE_CHECKING:
    from ludwig.models.base import BaseModel
    from ludwig.schema.model_types.base import ModelConfig


@DeveloperAPI
def get_base_type_registry() -> Dict:
    return {
        TEXT: TextFeatureMixin,
        CATEGORY: CategoryFeatureMixin,
        SET: SetFeatureMixin,
        BAG: BagFeatureMixin,
        BINARY: BinaryFeatureMixin,
        NUMBER: NumberFeatureMixin,
        SEQUENCE: SequenceFeatureMixin,
        TIMESERIES: TimeseriesFeatureMixin,
        IMAGE: ImageFeatureMixin,
        AUDIO: AudioFeatureMixin,
        H3: H3FeatureMixin,
        DATE: DateFeatureMixin,
        VECTOR: VectorFeatureMixin,
    }


@DeveloperAPI
def get_input_type_registry() -> Dict:
    return {
        TEXT: TextInputFeature,
        NUMBER: NumberInputFeature,
        BINARY: BinaryInputFeature,
        CATEGORY: CategoryInputFeature,
        SET: SetInputFeature,
        SEQUENCE: SequenceInputFeature,
        IMAGE: ImageInputFeature,
        AUDIO: AudioInputFeature,
        TIMESERIES: TimeseriesInputFeature,
        BAG: BagInputFeature,
        H3: H3InputFeature,
        DATE: DateInputFeature,
        VECTOR: VectorInputFeature,
    }


@DeveloperAPI
def get_output_type_registry() -> Dict:
    return {
        CATEGORY: CategoryOutputFeature,
        BINARY: BinaryOutputFeature,
        NUMBER: NumberOutputFeature,
        SEQUENCE: SequenceOutputFeature,
        SET: SetOutputFeature,
        TEXT: TextOutputFeature,
        VECTOR: VectorOutputFeature,
    }


def update_config_with_metadata(config_obj: "ModelConfig", training_set_metadata: Dict[str, Any]):
    # populate input features fields depending on data
    for input_feature in config_obj.input_features:
        feature = get_from_registry(input_feature.type, get_input_type_registry())
        feature.update_config_with_metadata(input_feature, training_set_metadata[input_feature.name])

    # populate output features fields depending on data
    for output_feature in config_obj.output_features:
        feature = get_from_registry(output_feature.type, get_output_type_registry())
        feature.update_config_with_metadata(output_feature, training_set_metadata[output_feature.name])


def update_config_with_model(config_obj: "ModelConfig", model: "BaseModel"):
    """Updates the config with the final input feature params given a model.

    This function should only be called to update the config after the model is initialized. Currently only implemented
    for input features because it is only relevant for HuggingFace text encoders. HuggingFace text encoders only know
    their final config after class initialization.
    """
    for input_feature in config_obj.input_features:
        model_input_feature = model.input_features.get(input_feature.name)
        model_input_feature.update_config_after_module_init(input_feature)



import importlib.util
import os
import re
import shutil
import tempfile
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

import pandas as pd
import torch

from ludwig.api import LudwigModel
from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import (
    AUDIO,
    BAG,
    BINARY,
    CATEGORY,
    DATE,
    IMAGE,
    INPUT_FEATURES,
    POSTPROCESSOR,
    PREDICTOR,
    PREPROCESSOR,
    SEQUENCE,
    SET,
    TEXT,
    TIMESERIES,
    TYPE,
    VECTOR,
)
from ludwig.data.dataset_synthesizer import build_synthetic_dataset
from ludwig.models.inference import (
    _InferencePostprocessor,
    _InferencePredictor,
    _InferencePreprocessor,
    InferenceModule,
)
from ludwig.types import ModelConfigDict
from ludwig.utils.inference_utils import to_inference_module_input_from_dataframe
from ludwig.utils.misc_utils import remove_empty_lines
from ludwig.utils.torch_utils import model_size, place_on_device
from ludwig.utils.types import TorchAudioTuple, TorchscriptPreprocessingInput

FEATURES_TO_CAST_AS_STRINGS = {BINARY, CATEGORY, BAG, SET, TEXT, SEQUENCE, TIMESERIES, VECTOR}

INFERENCE_STAGES = [PREPROCESSOR, PREDICTOR, POSTPROCESSOR]
INPUT = "INPUT"
OUTPUT = "OUTPUT"
ENSEMBLE = "ensemble"

INFERENCE_MODULE_TEMPLATE = """
from typing import Any, Dict, List, Union
import torch
from ludwig.utils.types import TorchscriptPreprocessingInput

class GeneratedInferenceModule(torch.nn.Module):
    def __init__(self, inference_module):
        super().__init__()
        self.inference_module = inference_module

    def forward(self, {input_signature}):
        with torch.no_grad():
            inputs: Dict[str, {input_type}] = {input_dict}
            results = self.inference_module(inputs)
            return {output_tuple}
"""

FEATURE_RESHAPE_SPEC = """reshape: {{ shape: [ {reshape_dims} ] }}
"""

TRITON_SPEC = """
    {{
        name: "{key}"
        data_type: {data_type}
        dims: [ {data_dims} ]
        {reshape_spec}
    }}"""

INSTANCE_SPEC = """
    {{
        count: {count}
        kind: {kind}
    }}"""

DYNAMIC_BATCHING_TEMPLATE = """dynamic_batching {{
    max_queue_delay_microseconds: {delay}
}}"""

TRITON_CONFIG_TEMPLATE = """name: "{model_name}"
platform: "pytorch_libtorch"
max_batch_size: {max_batch_size}
{dynamic_batching_spec}
input [{input_spec}
]
output [{output_spec}
]
instance_group [{instance_spec}
]
"""

ENSEMBLE_SCHEDULING_INPUT_MAP = """
      input_map {{
        key: "{key}"
        value: "{value}"
      }}"""
ENSEMBLE_SCHEDULING_OUTPUT_MAP = """
      output_map {{
        key: "{key}"
        value: "{value}"
      }}"""

ENSEMBLE_SCHEDULING_STEP = """
    {{
      model_name: "{ensemble_model_name}"
      model_version: -1
      {input_maps}
      {output_maps}
    }}"""

TRITON_ENSEMBLE_CONFIG_TEMPLATE = """name: "{model_name}"
platform: "ensemble"
max_batch_size: 0
input [{input_spec}
]
output [{output_spec}
]
ensemble_scheduling {{
  step [{ensemble_scheduling_steps}
  ]
}}
"""


def _get_type_map(dtype: str) -> str:
    """Return the Triton API type mapped to numpy type."""
    # see: https://github.com/triton-inference-server/server/blob/main/docs/model_configuration.md
    return {
        "bool": "TYPE_BOOL",
        "uint8": "TYPE_UINT8",
        "uint16": "TYPE_UINT16",
        "uint32": "TYPE_UINT32",
        "uint64": "TYPE_UINT64",
        "int8": "TYPE_INT8",
        "int16": "TYPE_INT16",
        "int32": "TYPE_INT32",
        "int64": "TYPE_INT64",
        "float16": "TYPE_FP16",
        "float32": "TYPE_FP32",
        "float64": "TYPE_FP64",
        "string": "TYPE_STRING",
        "torch.float32": "TYPE_FP32",
        "torch.float": "TYPE_FP32",
        "torch.float64": "TYPE_FP64",
        "torch.double": "TYPE_FP64",
        "torch.float16": "TYPE_FP16",
        "torch.half": "TYPE_FP16",
        "torch.uint8": "TYPE_UINT8",
        "torch.int8": "TYPE_INT8",
        "torch.int16": "TYPE_INT16",
        "torch.short": "TYPE_INT16",
        "torch.int32": "TYPE_INT32",
        "torch.int": "TYPE_INT32",
        "torch.int64": "TYPE_INT64",
        "torch.long": "TYPE_INT64",
        "torch.bool": "TYPE_BOOL",
    }[dtype]


def to_triton_dimension(content: Union[List[str], List[torch.Tensor], List[TorchAudioTuple], torch.Tensor]):
    # todo (Wael): tests for all types.
    if isinstance(content, list) and content:
        if isinstance(content[0], str):
            return [len(content)]
    elif isinstance(content, torch.Tensor):
        return list(content.size())
    return [-1]


def to_triton_type(content: Union[List[str], List[torch.Tensor], List[TorchAudioTuple], torch.Tensor]):
    # todo (Wael): tests for all types.
    if isinstance(content, list) and content:
        if isinstance(content[0], str):
            return _get_type_map("string")
    elif isinstance(content, torch.Tensor):
        return _get_type_map(str(content.dtype))


@DeveloperAPI
@dataclass
class TritonArtifact:
    """Dataclass for exported Triton artifacts."""

    # Name of the model.
    model_name: str

    # Model version.
    model_version: Union[int, str]

    # Triton backend (e.g. "pytorch_libtorch").
    platform: str

    # Model path.
    path: str

    # Type of artifact (application/octet-stream, plain text, etc.)
    content_type: str

    # Size of the artifact in bytes.
    content_length: int


@DeveloperAPI
@dataclass
class TritonConfigFeature:
    """Represents an input/output feature in a Triton config."""

    # Name of the feature.
    name: str

    # Ludwig type of the feature, or "tensor"
    ludwig_type: str

    # The data contents of the feature.
    content: Union[TorchscriptPreprocessingInput, torch.Tensor]

    # One of PREPROCESSOR, PREDICTOR, POSTPROCESSOR.
    inference_stage: str

    # One of INPUT, OUTPUT.
    kind: str

    # Index of the feature in the Triton Config.
    index: int

    def __post_init__(self):
        # removing non-alphanumeric characters as this will go in the wrapper function header.
        self.wrapper_signature_name = re.sub(r"[\W]+", "_", self.name)
        # get Triton type
        self.type = to_triton_type(self.content)
        # get dimension
        self.dimension = to_triton_dimension(self.content)
        # get ensemble_scheduling output_map key (same as "name" in input/output)
        self.key = f"{self.kind}__{self.index}"
        self.value = self._get_feature_ensemble_value()

    def _get_feature_ensemble_value(self) -> str:
        # get ensemble_scheduling output_map value.
        if self.inference_stage == PREPROCESSOR and self.kind == INPUT:
            return self.name
        if self.inference_stage == PREDICTOR and self.kind == INPUT:
            # PREPROCESSOR outputs and PREDICTOR inputs must have the same "value" attribute.
            return f"{PREPROCESSOR}_{OUTPUT}_{self.index}"
        elif self.inference_stage == POSTPROCESSOR and self.kind == INPUT:
            # PREDICTOR outputs and POSTPROCESSOR inputs must have the same "value" attribute.
            return f"{PREDICTOR}_{OUTPUT}_{self.index}"
        elif self.inference_stage == POSTPROCESSOR and self.kind == OUTPUT:
            return self.name
        else:
            return f"{self.inference_stage}_{self.kind}_{self.index}"

    def _get_wrapper_signature_type(self) -> str:
        if self.ludwig_type in FEATURES_TO_CAST_AS_STRINGS:
            return "List[str]"
        elif self.ludwig_type in [IMAGE, AUDIO, DATE, "tensor"]:
            return {
                IMAGE: "List[torch.Tensor]",
                AUDIO: "TorchAudioTuple",
                DATE: "List[torch.Tensor]",
                "tensor": "torch.Tensor",
            }[self.ludwig_type]
        return "torch.Tensor"


@DeveloperAPI
@dataclass
class TritonMaster:
    """Provides access to the Triton Config and the scripted module."""

    # The inference module.
    module: Union[_InferencePreprocessor, _InferencePredictor, _InferencePostprocessor]

    # An input for the module that will help determine the input and output dimensions.
    input_data_example: Dict[str, Union[TorchscriptPreprocessingInput, torch.Tensor]]

    # One of PREPROCESSOR, PREDICTOR, POSTPROCESSOR.
    inference_stage: str

    # Max batch size of the model. (Triton config param).
    max_batch_size: int

    # Max time a request to the Triton model spends in the queue. (Triton config param).
    max_queue_delay_microseconds: int

    # Name of the model as specified by the caller of the triton_export function.
    model_name: str

    # Base output directory. Corresponds to the Triton model registry.
    output_path: str

    # Triton model version.
    model_version: Union[int, str]

    # Ludwig config.
    ludwig_config: ModelConfigDict

    # One of "cpu", "cuda".
    device: str

    # Number of model instances on device.
    model_instance_count: int

    def __post_init__(self):
        """Extract input and output features and necessary information for a Triton config."""
        if self.inference_stage not in INFERENCE_STAGES:
            raise ValueError(f"Invalid inference stage. Choose one of {INFERENCE_STAGES}")

        self.full_model_name = self.model_name + "_" + self.inference_stage
        self.base_path = os.path.join(self.output_path, self.full_model_name)
        os.makedirs(self.base_path, exist_ok=True)

        self.output_data_example: Dict[str, Union[TorchscriptPreprocessingInput, torch.Tensor]] = self.module(
            self.input_data_example
        )

        # generate input and output features.
        self.input_features: List[TritonConfigFeature] = []
        for i, (feature_name, content) in enumerate(self.input_data_example.items()):
            ludwig_type = "tensor"
            if self.inference_stage == PREPROCESSOR:
                ludwig_type = self.ludwig_config[INPUT_FEATURES][i][TYPE]
            self.input_features.append(
                TritonConfigFeature(feature_name, ludwig_type, content, self.inference_stage, INPUT, i)
            )

        self.output_features: List[TritonConfigFeature] = []
        for i, (feature_name, content) in enumerate(self.output_data_example.items()):
            ludwig_type = "tensor"
            self.output_features.append(
                TritonConfigFeature(feature_name, ludwig_type, content, self.inference_stage, OUTPUT, i)
            )

    def save_model(self) -> TritonArtifact:
        """Script the model and saves it.

        Return the appropriate artifact.
        """
        if isinstance(self.model_version, str) and self.model_version.isdigit():
            self.model_version = int(self.model_version)
        if not isinstance(self.model_version, int) or self.model_version < 1:
            raise ValueError("Model version has to be a non-zero positive integer")
        pass

        # wrapper.py is optional and is just for visualizing the inputs/outputs to the model exported to Triton.
        wrapper_definition = TritonModel(
            self.module, self.input_features, self.output_features, self.inference_stage
        ).generate_inference_module_wrapper()
        with open(os.path.join(self.base_path, "wrapper.py"), "w") as f:
            f.write(wrapper_definition)

        os.makedirs(os.path.join(self.base_path, str(self.model_version)), exist_ok=True)
        model_path = os.path.join(self.base_path, str(self.model_version), "model.pt")

        self.model_ts = TritonModel(
            self.module, self.input_features, self.output_features, self.inference_stage
        ).generate_scripted_module()
        self.model_ts.save(model_path)

        model_artifact = TritonArtifact(
            model_name=self.full_model_name,
            model_version=self.model_version,
            platform="pytorch_libtorch",
            path=model_path,
            content_type="application/octet-stream",
            content_length=model_size(self.model_ts),
        )

        return model_artifact

    def save_config(self) -> TritonArtifact:
        """Save the Triton config.

        Return the appropriate artifact.
        """
        device = self.device
        if self.inference_stage != PREDICTOR:
            device = "cpu"
        self.config = TritonConfig(
            self.full_model_name,
            self.input_features,
            self.output_features,
            self.max_batch_size,
            self.max_queue_delay_microseconds,
            device,
            self.model_instance_count,
            self.inference_stage,
        )

        config_path = os.path.join(self.base_path, "config.pbtxt")
        with open(config_path, "w") as f:
            formatted_config = remove_empty_lines(self.config.get_model_config())
            f.write(formatted_config)

        config_artifact = TritonArtifact(
            model_name=self.full_model_name,
            model_version=self.model_version,
            platform="pytorch_libtorch",
            path=config_path,
            content_type="text/x-protobuf",
            content_length=os.path.getsize(config_path),
        )

        return config_artifact


@DeveloperAPI
@dataclass
class TritonEnsembleConfig:
    """Dataclass for creating and saving the Triton ensemble config."""

    # TritonMaster object for the preprocessor.
    triton_master_preprocessor: TritonMaster

    # TritonMaster object for the predictor.
    triton_master_predictor: TritonMaster

    # TritonMaster object for the postprocessor.
    triton_master_postprocessor: TritonMaster

    # Name of the model as specified by the caller of the triton_export function.
    model_name: str

    # Base output directory. Corresponds to the Triton model registry.
    output_path: str

    # Triton model version.
    model_version: Union[int, str]

    def __post_init__(self):
        self.ensemble_model_name = self.model_name
        self.base_path = os.path.join(self.output_path, self.ensemble_model_name)
        os.makedirs(self.base_path, exist_ok=True)

    def _get_ensemble_scheduling_input_maps(self, triton_features: List[TritonConfigFeature]) -> str:
        return "".join(
            ENSEMBLE_SCHEDULING_INPUT_MAP.format(key=feature.key, value=feature.value) for feature in triton_features
        )

    def _get_ensemble_scheduling_output_maps(self, triton_features: List[TritonConfigFeature]) -> str:
        return "".join(
            ENSEMBLE_SCHEDULING_OUTPUT_MAP.format(key=feature.key, value=feature.value) for feature in triton_features
        )

    def _get_ensemble_scheduling_step(self, triton_master: TritonMaster) -> str:
        return ENSEMBLE_SCHEDULING_STEP.format(
            ensemble_model_name=triton_master.config.full_model_name,
            input_maps=self._get_ensemble_scheduling_input_maps(triton_master.input_features),
            output_maps=self._get_ensemble_scheduling_output_maps(triton_master.output_features),
        )

    def _get_ensemble_spec(self, triton_features: List[TritonConfigFeature]) -> str:
        spec = []
        for feature in triton_features:
            spec.append(
                TRITON_SPEC.format(
                    key=feature.value,
                    data_type=feature.type,
                    data_dims=", ".join(str(dim) for dim in feature.dimension),  # check correctness
                    reshape_spec="",
                )
            )
        return ",".join(spec)

    def get_config(self) -> str:
        triton_masters = [
            self.triton_master_preprocessor,
            self.triton_master_predictor,
            self.triton_master_postprocessor,
        ]
        ensemble_scheduling_steps = ",".join(
            [self._get_ensemble_scheduling_step(triton_master) for triton_master in triton_masters]
        )
        return TRITON_ENSEMBLE_CONFIG_TEMPLATE.format(
            model_name=self.ensemble_model_name,
            input_spec=self._get_ensemble_spec(self.triton_master_preprocessor.input_features),
            output_spec=self._get_ensemble_spec(self.triton_master_postprocessor.output_features),
            ensemble_scheduling_steps=ensemble_scheduling_steps,
        )

    def save_ensemble_config(self) -> TritonArtifact:
        config_path = os.path.join(self.base_path, "config.pbtxt")
        with open(config_path, "w") as f:
            formatted_config = remove_empty_lines(self.get_config())
            f.write(formatted_config)

        config_artifact = TritonArtifact(
            model_name=self.ensemble_model_name,
            model_version=self.model_version,
            platform="ensemble",
            path=config_path,
            content_type="text/x-protobuf",
            content_length=os.path.getsize(config_path),
        )

        return config_artifact

    def save_ensemble_dummy_model(self) -> TritonArtifact:
        """Scripts the model and saves it."""
        if isinstance(self.model_version, str) and self.model_version.isdigit():
            self.model_version = int(self.model_version)
        if not isinstance(self.model_version, int) or self.model_version < 1:
            raise ValueError("Model version has to be a non-zero positive integer")
        pass

        os.makedirs(os.path.join(self.base_path, str(self.model_version)), exist_ok=True)
        model_path = os.path.join(self.base_path, str(self.model_version), "model.txt")
        with open(model_path, "w") as f:
            f.write("no model for the ensemble")

        model_artifact = TritonArtifact(
            model_name=self.ensemble_model_name,
            model_version=self.model_version,
            platform="ensemble",
            path=model_path,
            content_type="text/plain",
            content_length=os.path.getsize(model_path),
        )

        return model_artifact


@DeveloperAPI
@dataclass
class TritonConfig:
    """Enables the creation and export of a Triton config."""

    # Name of the model. Must be the same as the directory where the config is saved.
    full_model_name: str

    # Input features of the model.
    input_features: List[TritonConfigFeature]

    # Output features of the model.
    output_features: List[TritonConfigFeature]

    # Max batch size of the model. (Triton config param).
    max_batch_size: int

    # Max time a request to the Triton model spends in the queue. (Triton config param).
    max_queue_delay_microseconds: int

    # One of "cpu", "cuda".
    device: str

    # Number of model instances on device.
    model_instance_count: int

    # One of PREPROCESSOR, PREDICTOR, POSTPROCESSOR.
    inference_stage: str

    def _get_triton_spec(self, triton_features: List[TritonConfigFeature]) -> str:
        spec = []
        for feature in triton_features:
            spec.append(
                TRITON_SPEC.format(
                    key=feature.key,
                    data_type=feature.type,
                    data_dims=", ".join(str(dim) for dim in feature.dimension),  # check correctness
                    reshape_spec=self._get_reshape_spec(feature),
                )
            )
        return ",".join(spec)

    def _get_reshape_spec(self, feature) -> str:
        if feature.kind == INPUT and self.inference_stage == PREDICTOR:
            return FEATURE_RESHAPE_SPEC.format(reshape_dims=", ".join(str(dim) for dim in feature.dimension[1:]))
        return ""

    def _get_instance_spec(self) -> str:
        if self.device == "cpu":
            kind = "KIND_CPU"
        else:
            kind = "KIND_GPU"
        spec = INSTANCE_SPEC.format(count=self.model_instance_count, kind=kind)
        return spec

    def _get_dynamic_batching_spec(self) -> str:
        if self.inference_stage == PREDICTOR:
            return DYNAMIC_BATCHING_TEMPLATE.format(delay=self.max_queue_delay_microseconds)
        return ""

    def get_model_config(self) -> str:
        """Generate a Triton config for a model from the input and output features."""
        max_batch_size = self.max_batch_size
        if self.inference_stage != PREDICTOR:
            max_batch_size = 0

        config = TRITON_CONFIG_TEMPLATE.format(
            model_name=self.full_model_name,
            max_batch_size=max_batch_size,
            dynamic_batching_spec=self._get_dynamic_batching_spec(),
            input_spec=self._get_triton_spec(self.input_features),
            output_spec=self._get_triton_spec(self.output_features),
            instance_spec=self._get_instance_spec(),
        )
        return config


@DeveloperAPI
@dataclass
class TritonModel:
    """Enables the scripting and export of a model."""

    # The inference module.
    module: Union[_InferencePreprocessor, _InferencePredictor, _InferencePostprocessor]

    # Input features of the model.
    input_features: List[TritonConfigFeature]

    # Output features of the model.
    output_features: List[TritonConfigFeature]

    # One of PREPROCESSOR, PREDICTOR, POSTPROCESSOR.
    inference_stage: str

    def _get_dict_type_hint(self) -> str:
        return {
            PREPROCESSOR: "TorchscriptPreprocessingInput",
            PREDICTOR: "torch.Tensor",
            POSTPROCESSOR: "torch.Tensor",
        }[self.inference_stage]

    def _get_input_signature(self, triton_features: List[TritonConfigFeature]) -> str:
        elems = [
            f"{feature.wrapper_signature_name}: {feature._get_wrapper_signature_type()}" for feature in triton_features
        ]
        return ", ".join(elems)

    def _get_input_dict(self, triton_features: List[TritonConfigFeature]) -> str:
        elems = [f'"{feature.name}": {feature.wrapper_signature_name}' for feature in triton_features]
        return "{" + ", ".join(elems) + "}"

    def _get_output_tuple(self, triton_features: List[TritonConfigFeature]) -> str:
        elems = [f'results["{feature.name}"]' for feature in triton_features]
        return "(" + ", ".join(elems) + ",)"

    def generate_inference_module_wrapper(self) -> str:
        """Generate the class wrapper around an inference module."""
        return INFERENCE_MODULE_TEMPLATE.format(
            input_signature=self._get_input_signature(self.input_features),
            input_type=self._get_dict_type_hint(),
            input_dict=self._get_input_dict(self.input_features),
            output_tuple=self._get_output_tuple(self.output_features),
        )

    def generate_scripted_module(self):
        """Generate the scripted module from the wrapper class."""
        wrapper_definition = self.generate_inference_module_wrapper()
        with tempfile.TemporaryDirectory() as tmpdir:
            ts_path = os.path.join(tmpdir, "generated.py")
            with open(ts_path, "w") as f:
                f.write(wrapper_definition)

            spec = importlib.util.spec_from_file_location("generated.ts", ts_path)
            gen_ts = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(gen_ts)

            gen_module = gen_ts.GeneratedInferenceModule(self.module)
            scripted_module = torch.jit.script(gen_module)
        return scripted_module


def get_device_types_and_counts(
    preprocessor_num_instances: int,
    predictor_device_type: str,
    predictor_num_instances: int,
    postprocessor_num_instances: int,
) -> Tuple[List[str], List[int]]:
    """Retrun device types and instance counts for each of the three inference modules."""
    if predictor_device_type not in ["cuda", "cpu"]:
        raise ValueError('Invalid predictor device type. Choose one of ["cpu", "cuda"].')
    elif predictor_device_type == "cuda" and not torch.cuda.is_available():
        raise ValueError("Specified num_gpus > 0, but CUDA isn't available.")

    preprocessor_device_type = "cpu"
    postprocessor_device_type = "cpu"
    device_types = [preprocessor_device_type, predictor_device_type, postprocessor_device_type]
    device_counts = [preprocessor_num_instances, predictor_num_instances, postprocessor_num_instances]
    return device_types, device_counts


def get_inference_modules(model: LudwigModel, predictor_device_type: str) -> List[torch.jit.ScriptModule]:
    """Return the three inference modules."""
    inference_module = InferenceModule.from_ludwig_model(
        model.model, model.config, model.training_set_metadata, device=predictor_device_type
    )
    return [inference_module.preprocessor, inference_module.predictor, inference_module.postprocessor]


def get_example_input(
    model: LudwigModel, device_types: List[str], data_example: Union[None, pd.DataFrame]
) -> Dict[str, TorchscriptPreprocessingInput]:
    """Return an inference module-compatible input example.

    Generates a synthetic example if one is not provided.
    """
    config = model.config
    if data_example is None:
        features = config["input_features"] + config["output_features"]
        df = build_synthetic_dataset(dataset_size=1, features=features)
        data = [row for row in df]
        data_example = pd.DataFrame(data[1:], columns=data[0])
    return to_inference_module_input_from_dataframe(
        data_example.head(1), config, load_paths=True, device=device_types[0]
    )


def clean_up_synthetic_data():
    """Clean up synthetic example generated data for audio and image features."""
    shutil.rmtree("audio_files", ignore_errors=True)
    shutil.rmtree("image_files", ignore_errors=True)


@DeveloperAPI
def export_triton(
    model: LudwigModel,
    data_example: Optional[pd.DataFrame] = None,
    output_path: Optional[str] = "model_repository",
    model_name: Optional[str] = "ludwig_model",
    model_version: Optional[Union[int, str]] = 1,
    preprocessor_num_instances: Optional[int] = 1,
    predictor_device_type: Optional[str] = "cpu",
    predictor_num_instances: Optional[int] = 1,
    postprocessor_num_instances: Optional[int] = 1,
    predictor_max_batch_size: Optional[int] = 64,
    max_queue_delay_microseconds: Optional[int] = 100,
) -> List[TritonArtifact]:
    """Exports a torchscript model to a output path that serves as a repository for Triton Inference Server.

    # Inputs
    :param model: (LudwigModel) A ludwig model.
    :param data_example: (pd.DataFrame) an example from the dataset.
        Used to get dimensions throughout the pipeline.
    :param output_path: (str) The output path for the model repository.
    :param model_name: (str) The optional model name.
    :param model_version: (Union[int,str]) The optional model verison.
    :param preprocessor_num_instances: (int) number of instances for the preprocessor (on CPU).
    :param predictor_device_type: (str) device type for the predictor to be deployed on. One of "cpu" or "cuda"
    :param predictor_num_instances: (int) number of instances for the predictor.
    :param postprocessor_num_instances: (int) number of instances for the postprocessor (on CPU).
    :param predictor_max_batch_size: (int) max_batch_size parameter for the predictor Triton config.
    :param max_queue_delay_microseconds: (int) max_queue_delay_microseconds for all Triton configs.

    # Return
    :return: (List[TritonArtifact]) list of TritonArtifacts that contains information about exported artifacts.
    """

    device_types, instance_counts = get_device_types_and_counts(
        preprocessor_num_instances, predictor_device_type, predictor_num_instances, postprocessor_num_instances
    )
    split_modules = get_inference_modules(model, device_types[1])
    example_input = get_example_input(model, device_types, data_example)

    triton_masters = []
    triton_artifacts = []
    for i, module in enumerate(split_modules):
        example_input = place_on_device(example_input, device_types[i])
        triton_master = TritonMaster(
            module=module,
            input_data_example=example_input,
            inference_stage=INFERENCE_STAGES[i],
            max_batch_size=predictor_max_batch_size,
            max_queue_delay_microseconds=max_queue_delay_microseconds,
            model_name=model_name,
            output_path=output_path,
            model_version=model_version,
            ludwig_config=model.config,
            device=device_types[i],
            model_instance_count=instance_counts[i],
        )
        example_input = triton_master.output_data_example
        config_artifact = triton_master.save_config()
        model_artifact = triton_master.save_model()
        triton_masters.append(triton_master)
        triton_artifacts.extend([config_artifact, model_artifact])

    # saving ensemble config
    triton_master_preprocessor, triton_master_predictor, triton_master_postprocessor = triton_masters
    ensemble_config = TritonEnsembleConfig(
        triton_master_preprocessor,
        triton_master_predictor,
        triton_master_postprocessor,
        model_name,
        output_path,
        model_version,
    )
    config_artifact = ensemble_config.save_ensemble_config()
    model_artifact = ensemble_config.save_ensemble_dummy_model()
    triton_artifacts.extend([config_artifact, model_artifact])
    clean_up_synthetic_data()

    return triton_artifacts



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import re
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch

from ludwig.constants import NAME, PREPROCESSING, SEQUENCE, TEXT, TIMESERIES, TYPE
from ludwig.utils.data_utils import hash_dict
from ludwig.utils.strings_utils import get_tokenizer_from_registry, UNKNOWN_SYMBOL

SEQUENCE_TYPES = {SEQUENCE, TEXT, TIMESERIES}
FEATURE_NAME_SUFFIX = "__ludwig"
FEATURE_NAME_SUFFIX_LENGTH = len(FEATURE_NAME_SUFFIX)


def should_regularize(regularize_layers):
    regularize = False
    if isinstance(regularize_layers, bool) and regularize_layers:
        regularize = True
    elif isinstance(regularize_layers, (list, tuple)) and regularize_layers and regularize_layers[-1]:
        regularize = True
    return regularize


def set_str_to_idx(set_string, feature_dict, tokenizer_name):
    try:
        tokenizer = get_tokenizer_from_registry(tokenizer_name)()
    except ValueError:
        raise Exception(f"Tokenizer {tokenizer_name} not supported")

    out = [feature_dict.get(item, feature_dict[UNKNOWN_SYMBOL]) for item in tokenizer(set_string)]

    return np.array(out, dtype=np.int32)


def compute_token_probabilities(
    probabilities: Union[list, tuple, np.ndarray],
) -> np.ndarray:
    """Gets the maximum probability per timestep.

    Args:
        probabilities: An iterable of iterables or np.ndarray with shape (sequence_length, num_classes)
            where each inner iterable or np.ndarray is the probability distribution for a single timestep.
    Returns:
        An np.ndarray with shape (sequence_length,) containing the maximum probability for each timestep.
    """
    if isinstance(probabilities, (list, tuple)):
        max_probs = []
        for timestep_probs in probabilities:
            max_probs.append(np.max(timestep_probs))
        max_probs = np.array(max_probs)
    elif isinstance(probabilities, np.ndarray):
        max_probs = np.max(probabilities, axis=-1)
    else:
        raise ValueError(f"probabilities type must be in [list, tuple, np.ndarray]. Got {type(probabilities)}")
    return max_probs


def compute_sequence_probability(
    sequence_probabilities: np.ndarray,
    max_sequence_length: Optional[int] = None,
    return_log_prob: bool = True,
) -> float:
    """Computes the sequence level probability.

    Args:
        sequence_probabilities: An iterable of iterables or np.ndarray with shape (sequence_length,)
        max_sequence_length: The maximum sequence length to use. If None, uses the first dim of `sequence_probabilities`
        return_log_prob: Whether to return the log probability. Defaults to True.
    """
    if max_sequence_length is None:
        max_sequence_length = sequence_probabilities.shape[0]

    sequence_probabilities = sequence_probabilities[:max_sequence_length]

    if return_log_prob:
        return np.sum(np.log(sequence_probabilities))
    else:
        return np.prod(sequence_probabilities)


def sanitize(name):
    """Replaces invalid id characters."""
    return re.sub("\\W|^(?=\\d)", "_", name)


def compute_feature_hash(feature: dict) -> str:
    """This function computes a hash for each feature based on the preprocessing dictionary associated with each
    feature, as well as the feature's type.

    Args:
        feature: Feature dictionary

    Returns: Feature hash name
    """
    feature_data = dict(
        preprocessing=feature.get(PREPROCESSING, {}),
        type=feature[TYPE],
    )
    return sanitize(feature[NAME]) + "_" + hash_dict(feature_data).decode("ascii")


def get_input_size_with_dependencies(
    combiner_output_size: int, dependencies: List[str], other_output_features  # Dict[str, "OutputFeature"]
):
    """Returns the input size for the first layer of this output feature's FC stack, accounting for dependencies on
    other output features.

    In the forward pass, the hidden states of any dependent output features get concatenated with the combiner's output.
    If this output feature depends on other output features, then the input size for this feature's FCStack is the sum
    of the output sizes of other output features + the combiner's output size.
    """
    input_size_with_dependencies = combiner_output_size
    for feature_name in dependencies:
        if other_output_features[feature_name].fc_stack.num_layers:
            input_size_with_dependencies += other_output_features[feature_name].fc_stack.output_shape[-1]
        else:
            # 0-layer FCStack. Use the output feature's input size.
            input_size_with_dependencies += other_output_features[feature_name].input_size
    return input_size_with_dependencies


def get_module_dict_key_from_name(name: str, feature_name_suffix: str = FEATURE_NAME_SUFFIX) -> str:
    """Returns a key that's guaranteed to be compatible with torch."""
    key = name.replace(".", "__ludwig_punct_period__")
    return key + feature_name_suffix


def get_name_from_module_dict_key(key: str, feature_name_suffix_length: int = FEATURE_NAME_SUFFIX_LENGTH) -> str:
    """Reverse of get_module_dict_key_from_name."""
    name = key.replace("__ludwig_punct_period__", ".")
    return name[:-feature_name_suffix_length]


class LudwigFeatureDict(torch.nn.Module):
    """Torch ModuleDict wrapper that permits keys with any name.

    Torch's ModuleDict implementation doesn't allow certain keys to be used if they conflict with existing class
    attributes, e.g.

    > torch.nn.ModuleDict({'type': torch.nn.Module()})  # Raises KeyError.

    This class is a simple wrapper around torch's ModuleDict that mitigates possible conflicts by using a key-suffixing
    protocol.

    This is also tracked in Pytorch: https://github.com/pytorch/pytorch/issues/71203.
    """

    def __init__(self):
        super().__init__()
        self.module_dict = torch.nn.ModuleDict()
        self.internal_key_to_original_name_map = {}

    def get(self, key) -> torch.nn.Module:
        return self.module_dict[get_module_dict_key_from_name(key)]

    def set(self, key: str, module: torch.nn.Module) -> None:
        module_dict_key_name = get_module_dict_key_from_name(key)
        self.internal_key_to_original_name_map[module_dict_key_name] = key
        self.module_dict[module_dict_key_name] = module

    def __len__(self) -> int:
        return len(self.module_dict)

    def __next__(self) -> None:
        return next(iter(self))

    def __iter__(self) -> None:
        return iter(self.keys())

    def keys(self) -> List[str]:
        return [
            get_name_from_module_dict_key(feature_name)
            for feature_name in self.internal_key_to_original_name_map.keys()
        ]

    def values(self) -> List[torch.nn.Module]:
        return [module for _, module in self.module_dict.items()]

    def items(self) -> List[Tuple[str, torch.nn.Module]]:
        return [
            (get_name_from_module_dict_key(feature_name), module) for feature_name, module in self.module_dict.items()
        ]

    def update(self, modules: Dict[str, torch.nn.Module]) -> None:
        for feature_name, module in modules.items():
            self.set(feature_name, module)



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import copy
import logging

import yaml

from ludwig.api_annotations import DeveloperAPI
from ludwig.contrib import add_contrib_callback_args
from ludwig.features.feature_registries import get_input_type_registry
from ludwig.globals import LUDWIG_VERSION
from ludwig.schema.model_config import ModelConfig
from ludwig.schema.preprocessing import PreprocessingConfig
from ludwig.utils.backward_compatibility import upgrade_config_dict_to_latest_version
from ludwig.utils.data_utils import load_config_from_str, load_yaml
from ludwig.utils.fs_utils import open_file
from ludwig.utils.print_utils import print_ludwig

logger = logging.getLogger(__name__)

default_random_seed = 42

# Still needed for preprocessing  TODO(Connor): Refactor ludwig/data/preprocessing to use schema
default_feature_specific_preprocessing_parameters = {
    name: preproc_sect.get_schema_cls()().preprocessing.to_dict()
    for name, preproc_sect in get_input_type_registry().items()
}

default_training_preprocessing_parameters = copy.deepcopy(default_feature_specific_preprocessing_parameters)
default_training_preprocessing_parameters.update(PreprocessingConfig().to_dict())

default_prediction_preprocessing_parameters = copy.deepcopy(default_feature_specific_preprocessing_parameters)


@DeveloperAPI
def render_config(config=None, output=None, **kwargs):
    upgraded_config = upgrade_config_dict_to_latest_version(config)
    output_config = ModelConfig.from_dict(upgraded_config).to_dict()

    if output is None:
        print(yaml.safe_dump(output_config, None, sort_keys=False))
    else:
        with open_file(output, "w") as f:
            yaml.safe_dump(output_config, f, sort_keys=False)


@DeveloperAPI
def cli_render_config(sys_argv):
    parser = argparse.ArgumentParser(
        description="This script renders the full config from a user config.",
        prog="ludwig render_config",
        usage="%(prog)s [options]",
    )
    parser.add_argument(
        "-c",
        "--config",
        type=load_yaml,
        help="Path to the YAML file containing the model configuration",
    )
    parser.add_argument(
        "-cs",
        "--config_str",
        dest="config",
        type=load_config_from_str,
        help="JSON or YAML serialized string of the model configuration",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="output rendered YAML config path",
        required=False,
    )

    add_contrib_callback_args(parser)
    args = parser.parse_args(sys_argv)

    args.callbacks = args.callbacks or []
    for callback in args.callbacks:
        callback.on_cmdline("render_config", *sys_argv)

    print_ludwig("Render Config", LUDWIG_VERSION)
    render_config(**vars(args))



from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import utils as schema_utils


@DeveloperAPI
class BasePreprocessingConfig(schema_utils.BaseMarshmallowConfig):
    """Base class for input feature preprocessing. Not meant to be used directly.

    The dataclass format prevents arbitrary properties from being set. Consequently, in child classes, all properties
    from the corresponding input feature class are copied over: check each class to check which attributes are different
    from the preprocessing of each feature.
    """



from typing import Callable, Dict, List, Optional, Tuple, Union

import pytest
import torch

from ludwig.modules.convolutional_modules import (
    Conv1DLayer,
    Conv1DStack,
    Conv2DLayer,
    Conv2DLayerFixedPadding,
    Conv2DStack,
    ParallelConv1D,
    ParallelConv1DStack,
    ResNet,
    ResNetBlock,
    ResNetBlockLayer,
    ResNetBottleneckBlock,
)
from ludwig.utils.image_utils import get_img_output_shape
from tests.integration_tests.parameter_update_utils import check_module_parameters_updated

BATCH_SIZE = 2
SEQ_SIZE = 17
HIDDEN_SIZE = 8
NUM_FILTERS = 4

RANDOM_SEED = 1919


###
# Helper function to compute expected output shape
# for Conv1D related layers
###
def expected_seq_size(
    seq_size: int,  # input max sequence length
    padding: str,  # conv1d padding: 'same' or 'valid'
    kernel_size: int,  # conv1d kernel size
    stride: int,  # conv1d stride
    dilation: int,  # conv1d dilation rate
    pool_size: Union[None, int],  # pooling layer kernel size
    pool_padding: str,  # pooling layer padding: 'same' or 'valid'
    pool_stride: int,  # pooling layer stride
) -> int:
    # output shape for the convolutional layer
    output_seq_size = get_img_output_shape(
        img_height=0,  # img_height set to zero for 1D structure
        img_width=seq_size,  # img_width equates to max sequence length
        kernel_size=kernel_size,
        stride=stride,
        padding=padding,
        dilation=dilation,
    )
    if pool_size is not None:
        # pooling layer present, adjust expected output shape for pooling layer
        output_seq_size = get_img_output_shape(
            img_height=0,  # img_height set to zero for 1D structure
            img_width=output_seq_size[1],  # img_width equates to max sequence length
            kernel_size=pool_size,
            stride=pool_stride,
            padding=pool_padding,
            dilation=1,  # pooling layer only support unit dilation
        )
    return output_seq_size[1]


###
# 1D Convolutional Tests
###
@pytest.mark.parametrize("pool_function", ["max", "mean"])
@pytest.mark.parametrize(
    "pool_size, pool_padding, pool_stride",
    [(None, None, None), (3, "same", 1), (5, "same", 1), (3, "valid", 2), (5, "valid", 2)],
)
@pytest.mark.parametrize("dilation", [1, 2])
@pytest.mark.parametrize("strides, padding", [(1, "same"), (1, "valid"), (2, "valid")])
@pytest.mark.parametrize("kernel_size", [3, 5])
def test_conv1d_layer(
    kernel_size: int,
    strides: int,
    padding: str,
    dilation: int,
    pool_size: Union[None, int],
    pool_padding: str,
    pool_stride: int,
    pool_function: str,
) -> None:
    # make test repeatable
    torch.manual_seed(RANDOM_SEED)

    # setup synthetic tensor for test
    input = torch.randn([BATCH_SIZE, SEQ_SIZE, HIDDEN_SIZE], dtype=torch.float32)

    conv1_layer = Conv1DLayer(
        in_channels=HIDDEN_SIZE,
        out_channels=NUM_FILTERS,
        max_sequence_length=SEQ_SIZE,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        dilation=dilation,
        pool_function=pool_function,
        pool_size=pool_size,
        pool_strides=pool_stride,
        pool_padding=pool_padding,
    )

    out_tensor = conv1_layer(input)

    # check for correct output class
    assert isinstance(out_tensor, torch.Tensor)

    # check for correct output shape
    output_seq_size = expected_seq_size(
        seq_size=SEQ_SIZE,
        padding=padding,
        kernel_size=kernel_size,
        stride=strides,
        dilation=dilation,
        pool_size=pool_size,
        pool_padding=pool_padding,
        pool_stride=pool_stride,
    )
    assert out_tensor.size() == (BATCH_SIZE, output_seq_size, NUM_FILTERS)


@pytest.mark.parametrize("dropout", [0, 0.5])
@pytest.mark.parametrize(
    "layers, num_layers",
    [
        (None, None),  # setup up default number of layers with default values
        (None, 4),  # setup of 4 layers with default values
        ([{"num_filters": NUM_FILTERS - 2}, {"num_filters": NUM_FILTERS + 2}], None),  # 2 custom layers
    ],
)
def test_conv1d_stack(layers: Union[None, list], num_layers: Union[None, int], dropout: float) -> None:
    # make test repeatable
    torch.manual_seed(RANDOM_SEED)

    # setup synthetic input tensor for test
    input = torch.randn([BATCH_SIZE, SEQ_SIZE, HIDDEN_SIZE], dtype=torch.float32)

    conv1_stack = Conv1DStack(
        in_channels=HIDDEN_SIZE,
        out_channels=NUM_FILTERS,
        max_sequence_length=SEQ_SIZE,
        layers=layers,
        num_layers=num_layers,
        default_num_filters=NUM_FILTERS,
        default_dropout=dropout,
    )

    # check for correct stack formation
    if layers is None:
        assert len(conv1_stack.stack) == 6 if num_layers is None else num_layers
    else:
        # custom layer specification
        assert len(conv1_stack.stack) == len(layers)
        assert conv1_stack.stack[0].out_channels == NUM_FILTERS - 2
        assert conv1_stack.stack[1].out_channels == NUM_FILTERS + 2

    # generate output tensor
    out_tensor = conv1_stack(input)

    # check for correct output class
    assert isinstance(out_tensor, torch.Tensor)

    assert out_tensor.size()[1:] == conv1_stack.output_shape[:]

    # check for correct output shape
    last_module = conv1_stack.stack[-1]
    output_seq_size = expected_seq_size(
        seq_size=last_module.input_shape[0],
        padding=last_module.padding,
        kernel_size=last_module.kernel_size,
        stride=last_module.stride,
        dilation=last_module.dilation,
        pool_size=last_module.pool_size,
        pool_padding=last_module.pool_padding,
        pool_stride=last_module.pool_strides,
    )
    if layers is None:
        # default stack setup
        assert out_tensor.size() == (BATCH_SIZE, output_seq_size, NUM_FILTERS)
    else:
        # custom stack setup
        assert out_tensor.size() == (BATCH_SIZE, output_seq_size, NUM_FILTERS + 2)

    # check for parameter updates
    target = torch.randn(conv1_stack.output_shape)
    _, tpc, upc, not_updated = check_module_parameters_updated(conv1_stack, (input,), target)
    if dropout == 0:
        # all trainable parameters should be updated
        assert tpc == upc, (
            f"All parameter not updated. Parameters not updated: {not_updated}" f"\nModule structure:\n{conv1_stack}"
        )
    else:
        # with specified config and random seed, non-zero dropout update parameter count could take different values
        assert (tpc == upc) or (upc == 1), (
            f"All parameter not updated. Parameters not updated: {not_updated}" f"\nModule structure:\n{conv1_stack}"
        )


@pytest.mark.parametrize(
    "layers",
    [
        None,  # setup up default number of layers with default values
        [{"filter_size": 3}, {"filter_size": 4}],  # custom parallel layers
    ],
)
def test_parallel_conv1d(layers: Union[None, list]) -> None:
    input = torch.randn([BATCH_SIZE, SEQ_SIZE, HIDDEN_SIZE], dtype=torch.float32)

    parallel_conv1d = ParallelConv1D(
        in_channels=HIDDEN_SIZE,
        out_channels=NUM_FILTERS,
        max_sequence_length=SEQ_SIZE,
        layers=layers,
        default_num_filters=NUM_FILTERS,
    )

    # check for correct stack formation
    if layers is None:
        assert len(parallel_conv1d.parallel_layers) == 4
    else:
        # custom layer specification
        assert len(parallel_conv1d.parallel_layers) == len(layers)
        assert parallel_conv1d.parallel_layers[0].kernel_size == 3
        assert parallel_conv1d.parallel_layers[1].kernel_size == 4

    # generate output tensor
    out_tensor = parallel_conv1d(input)

    # check for correct output class
    assert isinstance(out_tensor, torch.Tensor)

    # check for correct output shape
    parallel_module = parallel_conv1d.parallel_layers[0]
    output_seq_size = expected_seq_size(
        seq_size=parallel_module.input_shape[0],
        padding=parallel_module.padding,
        kernel_size=parallel_module.kernel_size,
        stride=parallel_module.stride,
        dilation=parallel_module.dilation,
        pool_size=parallel_module.pool_size,
        pool_padding=parallel_module.pool_padding,
        pool_stride=parallel_module.pool_strides,
    )

    assert out_tensor.size() == (BATCH_SIZE, output_seq_size, len(parallel_conv1d.parallel_layers) * NUM_FILTERS)


TEST_FILTER_SIZE0 = 7
TEST_FILTER_SIZE1 = 5


@pytest.mark.parametrize("dropout", [0, 0.99])
@pytest.mark.parametrize(
    "stacked_layers",
    [
        None,  # setup up default number of layers with default values
        # custom stacked parallel layers
        [
            [  # parallel_conv1d_stack.stack[0]
                {"filter_size": 3},
                {"filter_size": 5},
                {"filter_size": TEST_FILTER_SIZE0},
            ],
            [  # parallel_conv1d_stack.stack[1]
                {"filter_size": 2},
                {"filter_size": 3},
                {"filter_size": 4},
                {"filter_size": TEST_FILTER_SIZE1},
            ],
        ],
    ],
)
def test_parallel_conv1d_stack(stacked_layers: Union[None, list], dropout: float) -> None:
    # make repeatable
    torch.manual_seed(RANDOM_SEED)

    # setup synthetic input tensor for test
    input = torch.randn([BATCH_SIZE, SEQ_SIZE, HIDDEN_SIZE], dtype=torch.float32)

    parallel_conv1d_stack = ParallelConv1DStack(
        in_channels=HIDDEN_SIZE,
        out_channels=NUM_FILTERS,
        max_sequence_length=SEQ_SIZE,
        stacked_layers=stacked_layers,
        default_num_filters=NUM_FILTERS,
        default_dropout=dropout,
    )

    # check for correct stack formation
    if stacked_layers is None:
        assert len(parallel_conv1d_stack.stack) == 3
        for i in range(len(parallel_conv1d_stack.stack)):
            assert len(parallel_conv1d_stack.stack[i].parallel_layers) == 4
    else:
        # spot check custom layer specification
        assert len(parallel_conv1d_stack.stack) == len(stacked_layers)
        assert len(parallel_conv1d_stack.stack[0].parallel_layers) == 3
        assert parallel_conv1d_stack.stack[0].parallel_layers[2].kernel_size == TEST_FILTER_SIZE0
        assert len(parallel_conv1d_stack.stack[1].parallel_layers) == 4
        assert parallel_conv1d_stack.stack[1].parallel_layers[3].kernel_size == TEST_FILTER_SIZE1

    # generate output tensor
    out_tensor = parallel_conv1d_stack(input)

    # check for correct output class
    assert isinstance(out_tensor, torch.Tensor)

    # check output shape
    assert out_tensor.size() == (BATCH_SIZE, *parallel_conv1d_stack.output_shape)

    # check for parameter updates
    target = torch.randn(parallel_conv1d_stack.output_shape)
    _, tpc, upc, not_updated = check_module_parameters_updated(parallel_conv1d_stack, (input,), target)
    if dropout == 0:
        # all trainable parameters should be updated
        assert tpc == upc, (
            f"All parameter not updated. Parameters not updated: {not_updated}"
            f"\nModule structure:\n{parallel_conv1d_stack}"
        )
    else:
        # with specified config and random seed, non-zero dropout update parameter count could take different values
        assert (tpc == upc) or (upc == 5), (
            f"All parameter not updated. Parameters not updated: {not_updated}"
            f"\nModule structure:\n{parallel_conv1d_stack}"
        )


###
#  2D Convolutional Tests
###
@pytest.mark.parametrize(
    ("img_height,img_width,in_channels,out_channels,pool_kernel_size," "pool_stride,pool_padding,pool_dilation"),
    [(224, 224, 3, 16, 2, 2, 0, 1)],
)
@pytest.mark.parametrize("stride,padding", [(1, "valid"), (1, "same"), (2, "valid")])
@pytest.mark.parametrize("kernel_size", [1, 3, 5])
@pytest.mark.parametrize("dilation", [1, 2])
@pytest.mark.parametrize("norm", ["batch", "layer"])
def test_conv2d_layer(
    img_height: int,
    img_width: int,
    in_channels: int,
    out_channels: int,
    kernel_size: int,
    stride: int,
    padding: Union[int, Tuple[int], str],
    dilation: Union[int, Tuple[int]],
    norm: str,
    pool_kernel_size: Union[int, Tuple[int]],
    pool_stride: int,
    pool_padding: Union[int, Tuple[int], str],
    pool_dilation: Union[int, Tuple[int]],
) -> None:
    conv2d_layer = Conv2DLayer(
        img_height=img_height,
        img_width=img_width,
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=kernel_size,
        stride=stride,
        padding=padding,
        dilation=dilation,
        norm=norm,
        pool_kernel_size=pool_kernel_size,
        pool_stride=pool_stride,
        pool_padding=pool_padding,
        pool_dilation=pool_dilation,
    )
    input_tensor = torch.rand(2, in_channels, img_height, img_width)
    output_tensor = conv2d_layer(input_tensor)
    assert output_tensor.shape[1:] == conv2d_layer.output_shape


@pytest.mark.parametrize("img_height,img_width", [(224, 224)])
@pytest.mark.parametrize(
    "layers,num_layers,first_in_channels",
    [
        (None, None, 3),
        (None, 5, 3),
        ([{"out_channels": 8}], None, 3),
        ([{"out_channels": 8, "in_channels": 3}], None, None),
    ],
)
def test_conv2d_stack(
    img_height: int,
    img_width: int,
    layers: Optional[List[Dict]],
    num_layers: Optional[int],
    first_in_channels: Optional[int],
) -> None:
    conv2d_stack = Conv2DStack(
        img_height=img_height,
        img_width=img_width,
        layers=layers,
        num_layers=num_layers,
        first_in_channels=first_in_channels,
    )
    input_tensor = torch.rand(2, 3, img_height, img_width)
    output_tensor = conv2d_stack(input_tensor)
    assert output_tensor.shape[1:] == conv2d_stack.output_shape


@pytest.mark.parametrize("img_height,img_width,in_channels", [(224, 224, 8)])
@pytest.mark.parametrize("stride", [1, 3])
@pytest.mark.parametrize("groups", [1, 8])
def test_conv2d_layer_fixed_padding(
    img_height: int, img_width: int, in_channels: int, stride: int, groups: int
) -> None:
    conv2d_fixed_padding = Conv2DLayerFixedPadding(
        img_height=img_height, img_width=img_width, in_channels=in_channels, stride=stride, groups=groups
    )
    input_tensor = torch.rand(2, in_channels, img_height, img_width)
    output_tensor = conv2d_fixed_padding(input_tensor)
    assert output_tensor.shape[1:] == conv2d_fixed_padding.output_shape


@pytest.mark.parametrize("img_height,img_width,first_in_channels,out_channels", [(224, 224, 64, 64)])
@pytest.mark.parametrize(
    "projection_shortcut",
    [None, Conv2DLayerFixedPadding(img_height=224, img_width=224, in_channels=64, out_channels=64)],
)
def test_resnet_block(
    img_height: int, img_width: int, first_in_channels: int, out_channels: int, projection_shortcut: Callable
) -> None:
    resnet_block = ResNetBlock(
        img_height=img_height,
        img_width=img_width,
        first_in_channels=first_in_channels,
        out_channels=out_channels,
        projection_shortcut=projection_shortcut,
    )
    input_tensor = torch.rand(2, first_in_channels, img_height, img_width)
    output_tensor = resnet_block(input_tensor)
    assert output_tensor.shape[1:] == resnet_block.output_shape


@pytest.mark.parametrize("img_height,img_width,first_in_channels,out_channels", [(224, 224, 64, 64)])
@pytest.mark.parametrize(
    "projection_shortcut",
    [None, Conv2DLayerFixedPadding(img_height=224, img_width=224, in_channels=64, out_channels=256)],
)
def test_resnet_bottleneck_block(
    img_height: int, img_width: int, first_in_channels: int, out_channels: int, projection_shortcut: Callable
) -> None:
    resnet_block = ResNetBottleneckBlock(
        img_height=img_height,
        img_width=img_width,
        first_in_channels=first_in_channels,
        out_channels=out_channels,
        projection_shortcut=projection_shortcut,
    )
    input_tensor = torch.rand(2, first_in_channels, img_height, img_width)
    output_tensor = resnet_block(input_tensor)
    assert output_tensor.shape[1:] == resnet_block.output_shape


@pytest.mark.parametrize("img_height,img_width,first_in_channels,out_channels,num_blocks", [(224, 224, 3, 32, 3)])
@pytest.mark.parametrize("is_bottleneck, block_fn", [(True, ResNetBottleneckBlock), (False, ResNetBlock)])
def test_resnet_block_layer(
    img_height: int,
    img_width: int,
    first_in_channels: int,
    out_channels: int,
    is_bottleneck: bool,
    block_fn: Union[ResNetBlock, ResNetBottleneckBlock],
    num_blocks: int,
):
    resnet_block_layer = ResNetBlockLayer(
        img_height=img_height,
        img_width=img_width,
        first_in_channels=first_in_channels,
        out_channels=out_channels,
        is_bottleneck=is_bottleneck,
        block_fn=block_fn,
        num_blocks=num_blocks,
    )
    input_tensor = torch.rand(2, first_in_channels, img_height, img_width)
    output_tensor = resnet_block_layer(input_tensor)
    assert output_tensor.shape[1:] == resnet_block_layer.output_shape


@pytest.mark.parametrize("img_height,img_width,first_in_channels,out_channels", [(224, 224, 3, 64)])
@pytest.mark.parametrize("resnet_size", [18, 34, 50])
def test_resnet(
    img_height: int,
    img_width: int,
    first_in_channels: int,
    out_channels: int,
    resnet_size: int,
):
    # make repeatable
    torch.manual_seed(RANDOM_SEED)

    resnet = ResNet(
        img_height=img_height,
        img_width=img_width,
        first_in_channels=first_in_channels,
        out_channels=out_channels,
        resnet_size=resnet_size,
    )
    input_tensor = torch.rand(2, first_in_channels, img_height, img_width)
    output_tensor = resnet(input_tensor)
    assert output_tensor.shape[1:] == resnet.output_shape

    # check for parameter updates
    target = torch.randn(output_tensor.shape)
    fpc, tpc, upc, not_updated = check_module_parameters_updated(resnet, (input_tensor,), target)
    # all trainable parameters should be updated
    assert tpc == upc, (
        f"All parameter not updated. Parameters not updated: {not_updated}" f"\nModule structure:\n{resnet}"
    )



from functools import lru_cache
from threading import Lock

import jsonschema.exceptions
from jsonschema import Draft7Validator, validate
from jsonschema.validators import extend

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import MODEL_ECD, MODEL_TYPE
from ludwig.error import ConfigValidationError

# TODO(travis): figure out why we need these imports to avoid circular import error
from ludwig.schema.combiners.utils import get_combiner_jsonschema  # noqa
from ludwig.schema.features.utils import get_input_feature_jsonschema, get_output_feature_jsonschema  # noqa
from ludwig.schema.hyperopt import get_hyperopt_jsonschema  # noqa
from ludwig.schema.preprocessing import get_preprocessing_jsonschema  # noqa
from ludwig.schema.trainer import get_model_type_jsonschema, get_trainer_jsonschema  # noqa
from ludwig.schema.utils import unload_jsonschema_from_marshmallow_class

VALIDATION_LOCK = Lock()


def get_ludwig_version_jsonschema():
    return {
        "type": "string",
        "title": "ludwig_version",
        "description": "Current Ludwig model schema version.",
    }


def get_backend_jsonschema():  # TODO: Is this still necessary after setting backend = Dict()?
    # TODO(travis): implement full backend schema
    return {
        "type": "object",
        "title": "backend",
        "description": "Backend configuration.",
        "additionalProperties": True,
    }


@DeveloperAPI
@lru_cache(maxsize=2)
def get_schema(model_type: str = MODEL_ECD):
    # Force populate combiner registry:
    import ludwig.combiners.combiners  # noqa: F401
    from ludwig.schema.model_types.base import model_type_schema_registry

    cls = model_type_schema_registry[model_type]
    props = unload_jsonschema_from_marshmallow_class(cls)["properties"]
    return {
        "type": "object",
        "properties": props,
        "title": "model_options",
        "description": "Settings for Ludwig configuration",
    }


@lru_cache(maxsize=2)
def get_validator():
    # Manually add support for tuples (pending upstream changes: https://github.com/Julian/jsonschema/issues/148):
    def custom_is_array(checker, instance):
        return isinstance(instance, list) or isinstance(instance, tuple)

    # This creates a new class, so cache to prevent a memory leak:
    # https://github.com/python-jsonschema/jsonschema/issues/868
    type_checker = Draft7Validator.TYPE_CHECKER.redefine("array", custom_is_array)
    return extend(Draft7Validator, type_checker=type_checker)


@DeveloperAPI
def check_schema(updated_config):
    """Emulates the pure JSONSchema validation that could be used in an environment without marshmallow.

    The incoming config may not be comprehensive, but is assumed to be up to date with the latest ludwig schema.
    """
    model_type = updated_config.get(MODEL_TYPE, MODEL_ECD)
    error = None
    with VALIDATION_LOCK:
        try:
            validate(instance=updated_config, schema=get_schema(model_type=model_type), cls=get_validator())
        except jsonschema.exceptions.ValidationError as e:
            # Capture error but don't raise here, otherwise we get the full output from `e`, which contains a dump
            # of the entire schema
            error = e

    if error is not None:
        raise ConfigValidationError(f"Failed to validate JSON schema for config. Error: {error.message}")



from typing import Optional

from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import utils as schema_utils
from ludwig.schema.combiners.base import BaseCombinerConfig
from ludwig.schema.metadata import COMBINER_METADATA
from ludwig.schema.utils import ludwig_dataclass

MAIN_SEQUENCE_FEATURE_DESCRIPTION = """
Name of a sequence, text, or time series feature to concatenate the outputs
of the other features to. If no `main_sequence_feature` is specified, the combiner will look through all the features in
the order they are defined in the configuration and will look for a feature with a rank 3 tensor output (sequence, text
or time series). If it cannot find one it will raise an exception, otherwise the output of that feature will be used for
concatenating the other features along the sequence `s` dimension. If there are other input features with a rank 3
output tensor, the combiner will concatenate them alongside the `s` dimension. All sequence-like input features must
have identical `s` dimension, otherwise an error will be thrown.
"""


@DeveloperAPI
@ludwig_dataclass
class SequenceConcatCombinerConfig(BaseCombinerConfig):
    """Parameters for sequence concat combiner."""

    @staticmethod
    def module_name():
        return "SequenceConcatCombiner"

    type: str = schema_utils.ProtectedString(
        "sequence_concat",
        description=COMBINER_METADATA["SequenceConcatCombiner"]["type"].long_description,
    )

    main_sequence_feature: Optional[str] = schema_utils.String(
        default=None,
        allow_none=True,
        description=MAIN_SEQUENCE_FEATURE_DESCRIPTION,
        parameter_metadata=COMBINER_METADATA["SequenceConcatCombiner"]["main_sequence_feature"],
    )

    reduce_output: Optional[str] = schema_utils.ReductionOptions(
        default=None,
        description="Strategy to use to aggregate the embeddings of the items of the set.",
        parameter_metadata=COMBINER_METADATA["SequenceConcatCombiner"]["reduce_output"],
    )



# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import shutil

import numpy as np
import torch

from ludwig.api import LudwigModel
from ludwig.collect import collect_activations, collect_weights, print_model_summary
from ludwig.constants import BATCH_SIZE, ENCODER, TRAINER, TYPE
from ludwig.utils.torch_utils import get_torch_device
from tests.integration_tests.utils import category_feature, ENCODERS, generate_data, sequence_feature

DEVICE = get_torch_device()


def _prepare_data(csv_filename):
    # Single sequence input, single category output
    input_features = [sequence_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": 2}, reduce_input="sum")]

    input_features[0][ENCODER][TYPE] = ENCODERS[0]

    # Generate test data
    data_csv = generate_data(input_features, output_features, csv_filename)
    return input_features, output_features, data_csv


def _train(input_features, output_features, data_csv, **kwargs):
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }

    model = LudwigModel(config)
    _, _, output_dir = model.train(dataset=data_csv, **kwargs)
    return model, output_dir


def _get_layers(model_path):
    model = LudwigModel.load(model_path)
    return [name for name, _ in model.model.named_children()]


def _collect_activations(model_path, layers, csv_filename, output_directory):
    return collect_activations(model_path, layers, dataset=csv_filename, output_directory=output_directory)


def test_collect_weights(tmpdir, csv_filename):
    output_dir = None
    try:
        model, output_dir = _train(*_prepare_data(csv_filename))
        model_path = os.path.join(output_dir, "model")

        # 1 for the encoder (embeddings).
        # 2 for the decoder classifier (w and b).
        weights = [w for _, w in model.model.collect_weights()]
        assert len(weights) == 3

        # Load model from disk to ensure correct weight names
        model_loaded = LudwigModel.load(model_path)
        tensor_names = [name for name, w in model_loaded.collect_weights()]
        assert len(tensor_names) == 3

        filenames = collect_weights(model_path, tensor_names, tmpdir)
        assert len(filenames) == 3

        for weight, filename in zip(weights, filenames):
            saved_weight = np.load(filename)
            assert torch.allclose(weight, torch.from_numpy(saved_weight).to(DEVICE), rtol=1.0e-4), filename
    finally:
        if output_dir:
            shutil.rmtree(output_dir, ignore_errors=True)


def test_collect_activations(tmpdir, csv_filename):
    output_dir = None
    try:
        model, output_dir = _train(*_prepare_data(csv_filename))
        model_path = os.path.join(output_dir, "model")

        # [last_hidden, logits, projection_input]
        filenames = _collect_activations(
            model_path, [name for name, _ in model.model.named_children()], csv_filename, tmpdir
        )
        assert len(filenames) == 3
    finally:
        if output_dir:
            shutil.rmtree(output_dir, ignore_errors=True)


def test_print_model_summary(csv_filename):
    output_dir = None
    model, output_dir = _train(*_prepare_data(csv_filename))
    model_path = os.path.join(output_dir, "model")
    print_model_summary(model_path)



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from typing import Any, Dict, List, Optional

import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import SET
from ludwig.encoders.base import Encoder
from ludwig.encoders.registry import register_encoder
from ludwig.modules.embedding_modules import EmbedSet
from ludwig.modules.fully_connected_modules import FCStack
from ludwig.schema.encoders.set_encoders import SetSparseEncoderConfig

logger = logging.getLogger(__name__)


@DeveloperAPI
@register_encoder("embed", SET)
class SetSparseEncoder(Encoder):
    def __init__(
        self,
        vocab: List[str],
        representation: str = "dense",
        embedding_size: int = 50,
        embeddings_trainable: bool = True,
        pretrained_embeddings: Optional[str] = None,
        embeddings_on_cpu: bool = False,
        fc_layers=None,
        num_fc_layers: int = 0,
        output_size: int = 10,
        use_bias: bool = True,
        weights_initializer: str = "xavier_uniform",
        bias_initializer: str = "zeros",
        norm: Optional[str] = None,
        norm_params: Optional[Dict[str, Any]] = None,
        activation: str = "relu",
        dropout: float = 0.0,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        self.vocab_size = len(vocab)

        logger.debug("  Embed")
        self.embed = EmbedSet(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )

        logger.debug("  FCStack")
        # TODO(shreya): Make sure this is updated when FCStack is updated
        self.fc_stack = FCStack(
            first_layer_input_size=self.embed.output_shape[-1],
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_output_size=output_size,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
        )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """
        Params:
            inputs: The inputs fed into the encoder.
                    Shape: [batch x vocab_size], type tf.int32.

        Returns:
            Embeddings of shape [batch x vocab_size x embed size], type float32.
        """
        hidden = self.embed(inputs)
        hidden = self.fc_stack(hidden)

        return hidden

    @staticmethod
    def get_schema_cls():
        return SetSparseEncoderConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.vocab_size])

    @property
    def output_shape(self) -> torch.Size:
        return self.fc_stack.output_shape



# !/usr/bin/env python
# Copyright (c) 2021 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from abc import ABC
from typing import Any, Callable, Dict, List, Union

from ludwig.api_annotations import PublicAPI
from ludwig.types import HyperoptConfigDict, ModelConfigDict, TrainingSetMetadataDict


@PublicAPI
class Callback(ABC):
    def on_cmdline(self, cmd: str, *args: List[str]):
        """Called when Ludwig is run on the command line with the callback enabled.

        :param cmd: The Ludwig subcommand being run, ex. "train", "evaluate", "predict", ...
        :param args: The full list of command-line arguments (sys.argv).
        """
        pass

    def on_preprocess_start(self, config: ModelConfigDict):
        """Called before preprocessing starts.

        :param config: The config dictionary.
        """
        pass

    def on_preprocess_end(self, training_set, validation_set, test_set, training_set_metadata: TrainingSetMetadataDict):
        """Called after preprocessing ends.

        :param training_set: The training set.
        :type training_set: ludwig.dataset.base.Dataset
        :param validation_set: The validation set.
        :type validation_set: ludwig.dataset.base.Dataset
        :param test_set: The test set.
        :type test_set: ludwig.dataset.base.Dataset
        :param training_set_metadata: Values inferred from the training set, including preprocessing settings,
                                      vocabularies, feature statistics, etc. Same as training_set_metadata.json.
        """

        pass

    def on_hyperopt_init(self, experiment_name: str):
        """Called to initialize state before hyperparameter optimization begins.

        :param experiment_name: The name of the current experiment.
        """
        pass

    def on_hyperopt_preprocessing_start(self, experiment_name: str):
        """Called before data preprocessing for hyperparameter optimization begins.

        :param experiment_name: The name of the current experiment.
        """
        pass

    def on_hyperopt_preprocessing_end(self, experiment_name: str):
        """Called after data preprocessing for hyperparameter optimization is completed.

        :param experiment_name: The name of the current experiment.
        """
        pass

    def on_hyperopt_start(self, experiment_name: str):
        """Called before any hyperparameter optimization trials are started.

        :param experiment_name: The name of the current experiment.
        """
        pass

    def on_hyperopt_end(self, experiment_name: str):
        """Called after all hyperparameter optimization trials are completed.

        :param experiment_name: The name of the current experiment.
        """
        pass

    def on_hyperopt_finish(self, experiment_name: str):
        """Deprecated.

        Use on_hyperopt_end instead.
        """
        # TODO(travis): remove in favor of on_hyperopt_end for naming consistency
        pass

    def on_hyperopt_trial_start(self, parameters: HyperoptConfigDict):
        """Called before the start of each hyperparameter optimization trial.

        :param parameters: The complete dictionary of parameters for this hyperparameter optimization experiment.
        """
        pass

    def on_hyperopt_trial_end(self, parameters: HyperoptConfigDict):
        """Called after the end of each hyperparameter optimization trial.

        :param parameters: The complete dictionary of parameters for this hyperparameter optimization experiment.
        """
        pass

    def should_stop_hyperopt(self):
        """Returns true if the entire hyperopt run (all trials) should be stopped.

        See: https://docs.ray.io/en/latest/tune/api_docs/stoppers.html#ray.tune.Stopper
        """
        return False

    def on_train_init(
        self,
        base_config: ModelConfigDict,
        experiment_directory: str,
        experiment_name: str,
        model_name: str,
        output_directory: str,
        resume_directory: Union[str, None],
    ):
        """Called after preprocessing, but before the creation of the model and trainer objects.

        :param base_config: The user-specified config, before the insertion of defaults or inferred values.
        :param experiment_directory: The experiment directory, same as output_directory if no experiment specified.
        :param experiment_name: The experiment name.
        :param model_name: The model name.
        :param output_directory: file path to where training results are stored.
        :param resume_directory: model directory to resume training from, or None.
        """
        pass

    def on_train_start(
        self,
        model,
        config: ModelConfigDict,
        config_fp: Union[str, None],
    ):
        """Called after creation of trainer, before the start of training.

        :param model: The ludwig model.
        :type model: ludwig.utils.torch_utils.LudwigModule
        :param config: The config dictionary.
        :param config_fp: The file path to the config, or none if config was passed to stdin.
        """
        pass

    def on_train_end(self, output_directory: str):
        """Called at the end of training, before the model is saved.

        :param output_directory: file path to where training results are stored.
        """
        pass

    def on_trainer_train_setup(self, trainer, save_path: str, is_coordinator: bool):
        """Called in every trainer (distributed or local) before training starts.

        :param trainer: The trainer instance.
        :type trainer: trainer: ludwig.models.Trainer
        :param save_path: The path to the directory model is saved in.
        :param is_coordinator: Is this trainer the coordinator.
        """
        pass

    def on_trainer_train_teardown(self, trainer, progress_tracker, save_path: str, is_coordinator: bool):
        """Called in every trainer (distributed or local) after training completes.

        :param trainer: The trainer instance.
        :type trainer: ludwig.models.trainer.Trainer
        :param progress_tracker: An object which tracks training progress.
        :type progress_tracker: ludwig.utils.trainer_utils.ProgressTracker
        :param save_path: The path to the directory model is saved in.
        :param is_coordinator: Is this trainer the coordinator.
        """
        pass

    def on_batch_start(self, trainer, progress_tracker, save_path: str):
        """Called on coordinator only before each batch.

        :param trainer: The trainer instance.
        :type trainer: ludwig.models.trainer.Trainer
        :param progress_tracker: An object which tracks training progress.
        :type progress_tracker: ludwig.utils.trainer_utils.ProgressTracker
        :param save_path: The path to the directory model is saved in.
        """
        pass

    def on_batch_end(self, trainer, progress_tracker, save_path: str):
        """Called on coordinator only after each batch.

        :param trainer: The trainer instance.
        :type trainer: ludwig.models.trainer.Trainer
        :param progress_tracker: An object which tracks training progress.
        :type progress_tracker: ludwig.utils.trainer_utils.ProgressTracker
        :param save_path: The path to the directory model is saved in.
        """
        pass

    def on_eval_start(self, trainer, progress_tracker, save_path: str):
        """Called on coordinator at the start of evaluation.

        :param trainer: The trainer instance.
        :type trainer: ludwig.models.trainer.Trainer
        :param progress_tracker: An object which tracks training progress.
        :type progress_tracker: ludwig.utils.trainer_utils.ProgressTracker
        :param save_path: The path to the directory model is saved in.
        """
        pass

    def on_eval_end(self, trainer, progress_tracker, save_path: str):
        """Called on coordinator at the end of evaluation.

        :param trainer: The trainer instance.
        :type trainer: ludwig.models.trainer.Trainer
        :param progress_tracker: An object which tracks training progress.
        :type progress_tracker: ludwig.utils.trainer_utils.ProgressTracker
        :param save_path: The path to the directory model is saved in.
        """
        pass

    def on_epoch_start(self, trainer, progress_tracker, save_path: str):
        """Called on coordinator only before the start of each epoch.

        :param trainer: The trainer instance.
        :type trainer: ludwig.models.trainer.Trainer
        :param progress_tracker: An object which tracks training progress.
        :type progress_tracker: ludwig.utils.trainer_utils.ProgressTracker
        :param save_path: The path to the directory model is saved in.
        """
        pass

    def on_epoch_end(self, trainer, progress_tracker, save_path: str):
        """Called on coordinator only after the end of each epoch.

        :param trainer: The trainer instance.
        :type trainer: ludwig.models.trainer.Trainer
        :param progress_tracker: An object which tracks training progress.
        :type progress_tracker: ludwig.utils.trainer_utils.ProgressTracker
        :param save_path: The path to the directory model is saved in.
        """
        pass

    def on_validation_start(self, trainer, progress_tracker, save_path: str):
        """Called on coordinator before validation starts.

        :param trainer: The trainer instance.
        :type trainer: ludwig.models.trainer.Trainer
        :param progress_tracker: An object which tracks training progress.
        :type progress_tracker: ludwig.utils.trainer_utils.ProgressTracker
        :param save_path: The path to the directory model is saved in.
        """
        pass

    def on_validation_end(self, trainer, progress_tracker, save_path: str):
        """Called on coordinator after validation is complete.

        :param trainer: The trainer instance.
        :type trainer: ludwig.models.trainer.Trainer
        :param progress_tracker: An object which tracks training progress.
        :type progress_tracker: ludwig.utils.trainer_utils.ProgressTracker
        :param save_path: The path to the directory model is saved in.
        """
        pass

    def on_test_start(self, trainer, progress_tracker, save_path: str):
        """Called on coordinator before testing starts.

        :param trainer: The trainer instance.
        :type trainer: ludwig.models.trainer.Trainer
        :param progress_tracker: An object which tracks training progress.
        :type progress_tracker: ludwig.utils.trainer_utils.ProgressTracker
        :param save_path: The path to the directory model is saved in.
        """
        pass

    def on_test_end(self, trainer, progress_tracker, save_path: str):
        """Called on coordinator after testing ends.

        :param trainer: The trainer instance.
        :type trainer: ludwig.models.trainer.Trainer
        :param progress_tracker: An object which tracks training progress.
        :type progress_tracker: ludwig.utils.trainer_utils.ProgressTracker
        :param save_path: The path to the directory model is saved in.
        """
        pass

    def should_early_stop(self, trainer, progress_tracker, is_coordinator):
        # Triggers early stopping if any callback on any worker returns True
        return False

    def on_build_metadata_start(self, df, mode: str):
        """Called before building metadata for dataset.

        :param df: The dataset.
        :type df: pd.DataFrame
        :param mode: "prediction", "training", or None.
        """
        pass

    def on_build_metadata_end(self, df, mode):
        """Called after building dataset metadata.

        :param df: The dataset.
        :type df: pd.DataFrame
        :param mode: "prediction", "training", or None.
        """
        pass

    def on_build_data_start(self, df, mode):
        """Called before build_data, which does preprocessing, handling missing values, adding metadata to
        training_set_metadata.

        :param df: The dataset.
        :type df: pd.DataFrame
        :param mode: "prediction", "training", or None.
        """
        pass

    def on_build_data_end(self, df, mode):
        """Called after build_data completes.

        :param df: The dataset.
        :type df: pd.DataFrame
        :param mode: "prediction", "training", or None.
        """
        pass

    def on_evaluation_start(self):
        """Called before preprocessing for evaluation."""
        pass

    def on_evaluation_end(self):
        """Called after evaluation is complete."""
        pass

    def on_visualize_figure(self, fig):
        """Called after a visualization is generated.

        :param fig: The figure.
        :type fig: matplotlib.figure.Figure
        """
        pass

    def on_ludwig_end(self):
        """Convenience method for any cleanup.

        Not yet implemented.
        """
        pass

    def prepare_ray_tune(self, train_fn: Callable, tune_config: Dict[str, Any], tune_callbacks: List[Callable]):
        """Configures Ray Tune callback and config.

        :param train_fn: The function which runs the experiment trial.
        :param tune_config: The ray tune configuration dictionary.
        :param tune_callbacks: List of callbacks (not used yet).

        :returns: Tuple[Callable, Dict] The train_fn and tune_config, which will be passed to ray tune.
        """
        return train_fn, tune_config
